# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Conditions for the Executor module.
"""

import functools
import types
from abc import ABC, abstractmethod
from enum import Enum, auto
from time import time as time_now
from typing import Any, Dict, Optional, Tuple

from texar.torch.run.executor_utils import MetricList
from texar.torch.utils.types import MaybeTuple

# pylint: disable=unused-argument

__all__ = [
    "Event",
    "EventPoint",
    "Condition",
    "epoch",
    "iteration",
    "validation",
    "consecutive",
    "time",
]


class Event(Enum):
    Iteration = auto()
    Epoch = auto()
    Training = auto()
    Validation = auto()
    ValidationIteration = auto()
    Testing = auto()
    TestingIteration = auto()
    ParameterUpdate = auto()


EventPoint = Tuple[Event, bool]


class Condition(ABC):
    _hooks: Dict[EventPoint, Any]

    @property
    @abstractmethod
    def _hash_attributes(self) -> MaybeTuple[Any]:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Condition):
            return False
        return self._hash_attributes == other._hash_attributes  # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._hash_attributes)

    @property
    def hooks(self) -> Dict[EventPoint, Any]:
        return self._hooks

    def __init__(self):
        self._hooks = {}
        for hook_name in dir(self):
            if not hook_name.startswith("check_"):
                continue
            name = hook_name
            if name.endswith("_begin"):
                name = name[6:-6]
                point = False
            elif name.endswith("_end"):
                name = name[6:-4]
                point = True
            else:
                raise ValueError(
                    "Final part of hook name must be 'begin' or 'end'")
            if name not in Event.__members__:
                name = ''.join(x.capitalize() for x in name.split("_"))
                if name not in Event.__members__:
                    raise ValueError(
                        f"Hook name '{hook_name}' is not a valid event")
            event = Event.__members__[name]
            self._hooks[(event, point)] = getattr(self, hook_name)


class epoch(Condition):
    r"""Triggers when the specified number of epochs has ended.

    Args:
        num_epochs (int): The number of epochs to wait before triggering the
            event. In other words, the event is triggered every
            :attr:`num_epochs` epochs.
    """

    def __init__(self, num_epochs: int = 1):
        if not isinstance(num_epochs, int) or num_epochs <= 0:
            raise ValueError("`num_epochs` must be a positive integer")
        super().__init__()
        self.num_epochs = num_epochs
        self.count = 0

    @property
    def _hash_attributes(self):
        return self.num_epochs

    def check_epoch_end(self, executor) -> bool:
        self.count += 1
        if self.count == self.num_epochs:
            self.count = 0
            return True
        return False


class iteration(Condition):
    r"""Triggers when the specified number of iterations had ended.

    Args:
        num_iters (int): The number of iterations to wait before triggering the
            event. In other words, the event is triggered every
            :attr:`num_iters` iterations.
        mode (str): The mode under which iterations are counted. Available
            choices are ``"train"``, ``"valid"``, and ``"test"``. Defaults to
            ``"train"``.
    """

    def __new__(cls, num_iters: int = 1, mode: str = "train"):
        obj = super().__new__(cls)
        # pylint: disable=protected-access
        if mode == "train":
            obj.check_iteration_end = obj._check_iteration_end
        elif mode == "valid":
            obj.check_validation_iteration_end = obj._check_iteration_end
        elif mode == "test":
            obj.check_testing_iteration_end = obj._check_iteration_end
        else:
            raise ValueError(f"Invalid mode {mode}")
        # pylint: enable=protected-access
        return obj

    def __init__(self, num_iters: int = 1, mode: str = "train"):
        if not isinstance(num_iters, int) or num_iters <= 0:
            raise ValueError("`num_iters` must be a positive integer")
        super().__init__()
        self.num_iters = num_iters
        self.count = 0

    @property
    def _hash_attributes(self):
        return self.num_iters

    def _check_iteration_end(self, executor) -> bool:
        self.count += 1
        if self.count == self.num_iters:
            self.count = 0
            return True
        return False


class validation(Condition):
    r"""Triggers when validation ends, and optionally checks if validation
    results improve or worsen.

    Args:
        num_validations (int): The number of validations to wait before
            triggering the event. In other words, the event is triggered every
            :attr:`num_validations` validations.
        better (bool, optional): If `True`, this event only triggers when
            validation results improve; if `False`, only triggers when results
            worsen. Defaults to `None`, in which case the event triggers
            regardless of results.
    """

    def __init__(self, num_validations: int = 1, better: Optional[bool] = None):
        if not isinstance(num_validations, int) or num_validations <= 0:
            raise ValueError("`num_validations` must be a positive integer")
        super().__init__()
        self.num_valids = num_validations
        self.count = 0
        self.better = better
        self.prev_result: Optional[MetricList] = None

    @property
    def _hash_attributes(self):
        return self.num_valids, self.better

    def check_validation_end(self, executor) -> bool:
        self.count += 1
        if self.count < self.num_valids:
            return False
        self.count = 0
        if self.better is None:
            return True
        metrics = executor.status["eval_metric"]
        cur_result = MetricList(metrics)
        if self.prev_result is not None:
            better = cur_result > self.prev_result
        else:
            better = True
        if better:
            self.prev_result = cur_result
        return better == self.better


class consecutive(Condition):
    r"""Triggers when the specified condition passes checks for several times
    consecutively.

    For example: :python:`consecutive(validation(better=False), times=3)` would
    trigger if validation results do not improve for 3 times in a row.

    .. warning::
        This method works by calling the inner condition at each event point
        that it registers. The consecutive counter is reset to zero if any check
        returns `False`. Thus, the behavior of :class:`consecutive` might be
        different to what you expect. For instance:

        - :python:`cond.consecutive(cond.iteration(1), n_times)` is equivalent
          to :python:`cond.iteration(n_times)`.
        - :python:`cond.consecutive(cond.iteration(2), n_times)` will never
          trigger.

        It is recommended against using :class:`consecutive` for conditions
        except :class:`validation`. You should also be careful when implementing
        custom conditions for using with :class:`consecutive`.

    .. warning::
        Conditions are stateful objects. Using a registered condition as the
        inner condition here could result in unexpected behaviors. For example:

        .. code-block:: python

            my_cond = cond.validation(better=True)
            executor.on(my_cond, some_action)
            executor.on(cond.consecutive(my_cond, 2), some_other_action)

        In the code above, if no other conditions are registered,
        :python:`some_other_action` will never be called. This is because both
        conditions are checked at the end of each iteration, but the
        :class:`consecutive` condition internally checks :python:`my_cond`,
        which has already updated the previous best result that it stored. As a
        result, the check will never succeed.

    Args:
        cond: The base condition to check.
        times (int): The number of times the base condition should pass checks
            consecutively.
        clear_after_trigger (bool): Whether the counter should be cleared after
            the event is triggered. If :attr:`clear_after_trigger` is set to
            `False`, once this event is triggered, it will trigger every time
            :attr:`cond` is triggered, until :attr:`cond` fails to trigger at
            some point. Defaults to `True`.
    """

    def __init__(self, cond: Condition, times: int,
                 clear_after_trigger: bool = True):
        super().__init__()
        self.cond = cond
        self.times = times
        self.count = 0
        self.clear_after_trigger = clear_after_trigger

        for hook_point, method in self.cond.hooks.items():
            self._hooks[hook_point] = self._create_check_method(method)

    @property
    def _hash_attributes(self):
        return self.cond, self.times, self.clear_after_trigger

    def _create_check_method(self, method):
        @functools.wraps(method)
        def check_fn(self, executor) -> bool:
            if method(executor):
                self.count += 1
                if self.count >= self.times:
                    if self.clear_after_trigger:
                        self.count = 0
                    return True
            else:
                self.count = 0
            return False

        return types.MethodType(check_fn, self)


class once(Condition):
    r"""Triggers only when the specified condition triggers for the first time.

    Internally, this condition calls the
    :meth:`~texar.torch.run.Executor.remove_action` method to remove itself from
    the registered actions.

    For example: :python:`once(iteration(5))` would only trigger on the 5th
    epoch of the entire training loop.

    .. warning::
        Conditions are stateful objects. Using a registered condition as the
        inner condition here could result in unexpected behaviors. Please refer
        to :class:`consecutive` for a concrete example.

    Args:
        cond: The base condition to check.
    """

    def __init__(self, cond: Condition):
        super().__init__()
        self.cond = cond

        for hook_point, method in self.cond.hooks.items():
            self._hooks[hook_point] = self._create_check_method(method)

    @property
    def _hash_attributes(self):
        return self.cond

    def _create_check_method(self, method):
        @functools.wraps(method)
        def check_fn(self, executor) -> bool:
            if method(executor):
                executor.remove_action()
                return True
            return False

        return types.MethodType(check_fn, self)


class time(Condition):
    def __init__(self, *, hours: Optional[float] = None,
                 minutes: Optional[float] = None,
                 seconds: Optional[float] = None,
                 only_training: bool = True):
        super().__init__()
        self.seconds = 0.0
        if hours is not None:
            self.seconds += hours * 3600.0
        if minutes is not None:
            self.seconds += minutes * 60.0
        if seconds is not None:
            self.seconds += seconds
        self.only_training = only_training
        self.start_time: Optional[float] = None
        self.accumulated_time = 0.0

    @property
    def _hash_attributes(self):
        return self.seconds, self.only_training

    def _should_trigger(self) -> bool:
        total_time = self.accumulated_time
        if self.start_time is None:
            cur_time = None
        else:
            cur_time = time_now()
            total_time += cur_time - self.start_time
        self.start_time = cur_time
        if total_time >= self.seconds:
            self.accumulated_time = 0.0
            return True
        else:
            self.accumulated_time = total_time
            return False

    def check_training_begin(self, executor) -> bool:
        self.start_time = time_now()
        return False

    def check_training_end(self, executor) -> bool:
        return self._should_trigger()

    def check_validation_begin(self, executor) -> bool:
        if self.only_training and self.start_time is not None:
            self.accumulated_time += time_now() - self.start_time
            self.start_time = None
        return self._should_trigger()

    def check_validation_end(self, executor) -> bool:
        if self.only_training:
            self.start_time = time_now()
            return False
        else:
            return self._should_trigger()

    def check_testing_begin(self, executor) -> bool:
        if self.only_training and self.start_time is not None:
            self.accumulated_time += time_now() - self.start_time
            self.start_time = None
        return self._should_trigger()

    def check_testing_end(self, executor) -> bool:
        if self.only_training:
            self.start_time = time_now()
            return False
        else:
            return self._should_trigger()

    def check_iteration_end(self, executor) -> bool:
        return self._should_trigger()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for XLNet model utils.
"""

import unittest

import torch

from texar.torch.modules.pretrained.xlnet_utils import *


class XLNetModelUtilsTest(unittest.TestCase):
    r"""Tests XLNet model utils.
    """

    def test_PositionWiseFF(self):

        # Case 1
        model = PositionWiseFF()
        inputs = torch.rand(32, model._hparams.hidden_dim)
        outputs = model(inputs)
        self.assertEqual(outputs.shape, torch.Size([32,
                                                    model._hparams.hidden_dim]))

        # Case 2
        hparams = {
            "hidden_dim": 16,
            "ffn_inner_dim": 32,
            "dropout": 0.1,
            "activation": 'relu',
        }
        model = PositionWiseFF(hparams=hparams)
        inputs = torch.rand(32, 16)
        outputs = model(inputs)
        self.assertEqual(outputs.shape, torch.Size([32, 16]))

        # Case 3
        hparams = {
            "hidden_dim": 16,
            "ffn_inner_dim": 32,
            "dropout": 0.1,
            "activation": 'gelu',
        }
        model = PositionWiseFF(hparams=hparams)
        inputs = torch.rand(32, 16)
        outputs = model(inputs)
        self.assertEqual(outputs.shape, torch.Size([32, 16]))

    def test_RelativeMultiheadAttention(self):

        model = RelativeMultiheadAttention()

        states_h = torch.rand(16, 32, model._hparams.hidden_dim)
        pos_embed = torch.rand(24, 32, model._hparams.hidden_dim)

        output_h, output_g = model(states_h=states_h, pos_embed=pos_embed)

        self.assertEqual(output_h.shape,
                         torch.Size([16, 32, model._hparams.hidden_dim]))
        self.assertEqual(output_g, None)

    def test_RelativePositionalEncoding(self):

        batch_size = 16
        seq_len = 8
        total_len = 32

        # Case 1
        model = RelativePositionalEncoding()
        pos_embed = model(batch_size=batch_size,
                          seq_len=seq_len,
                          total_len=total_len)
        self.assertEqual(pos_embed.shape,
                         torch.Size([40, 16, model._hparams.dim]))

        # Case 2
        model = RelativePositionalEncoding()
        pos_embed = model(batch_size=batch_size,
                          seq_len=seq_len,
                          total_len=total_len,
                          attn_type='uni')
        self.assertEqual(pos_embed.shape,
                         torch.Size([33, 16, model._hparams.dim]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for entropy.
"""

import unittest

import torch

from texar.torch.losses import entropy
from texar.torch.utils.shapes import get_rank


class EntropyTest(unittest.TestCase):
    """Tests entropy.
    """

    def setUp(self):
        self._batch_size = 64
        self._max_time = 128
        self._d = 16
        self._distribution_dim = 32
        self._logits = torch.rand(self._batch_size, self._d,
                                  self._distribution_dim)
        self._sequence_logits = torch.rand(
            self._batch_size, self._max_time, self._d,
            self._distribution_dim)
        self._sequence_length = torch.randint(
            high=self._max_time, size=(self._batch_size,))

    def _test_entropy(self, entropy_fn, logits, sequence_length=None):
        if sequence_length is None:
            entropy = entropy_fn(logits)
            rank = get_rank(entropy)
            self.assertEqual(rank, 0)

            entropy = entropy_fn(logits, average_across_batch=False)
            rank = get_rank(entropy)
            self.assertEqual(rank, 1)
            self.assertEqual(entropy.shape, torch.Size([self._batch_size]))
        else:
            entropy = entropy_fn(logits, sequence_length=sequence_length)
            rank = get_rank(entropy)
            self.assertEqual(rank, 0)

            entropy = entropy_fn(logits, sequence_length=sequence_length,
                                 sum_over_timesteps=False)
            rank = get_rank(entropy)
            self.assertEqual(rank, 1)
            self.assertEqual(entropy.shape, torch.Size([self._max_time]))

            entropy = entropy_fn(logits, sequence_length=sequence_length,
                                 sum_over_timesteps=False,
                                 average_across_timesteps=True,
                                 average_across_batch=False)
            rank = get_rank(entropy)
            self.assertEqual(rank, 1)
            self.assertEqual(entropy.shape, torch.Size([self._batch_size]))

            entropy = entropy_fn(logits, sequence_length=sequence_length,
                                 sum_over_timesteps=False,
                                 average_across_batch=False)
            rank = get_rank(entropy)
            self.assertEqual(rank, 2)
            self.assertEqual(entropy.shape, torch.Size([self._batch_size,
                                                        self._max_time]))

            sequence_length_time = torch.randint(size=(self._max_time,),
                                                 high=self._batch_size)
            entropy = entropy_fn(logits,
                                 sequence_length=sequence_length_time,
                                 sum_over_timesteps=False,
                                 average_across_batch=False,
                                 time_major=True)
            self.assertEqual(entropy.shape, torch.Size([self._batch_size,
                                                        self._max_time]))

    def test_entropy_with_logits(self):
        """Tests `entropy_with_logits`
        """
        self._test_entropy(
            entropy.entropy_with_logits, self._logits)

    def test_sequence_entropy_with_logits(self):
        """Tests `sequence_entropy_with_logits`
        """
        self._test_entropy(
            entropy.sequence_entropy_with_logits, self._sequence_logits,
            sequence_length=self._sequence_length)


if __name__ == "__main__":
    unittest.main()



import sys
import setuptools

long_description = """
Texar-PyTorch is an open-source toolkit based on PyTorch,
aiming to support a broad set of machine learning especially text generation
tasks, such as machine translation, dialog, summarization, content manipulation,
language modeling, and so on.

Texar is designed for both researchers and practitioners for fast prototyping
and experimentation. Checkout https://github.com/asyml/texar for the TensorFlow
version which has the same functionalities and (mostly) the same interfaces.
"""

if sys.version_info < (3, 6):
    sys.exit('Python>=3.6 is required by Texar-PyTorch.')

setuptools.setup(
    name="texar-pytorch",
    version="0.1.1",
    url="https://github.com/asyml/texar-pytorch",

    description="Toolkit for Machine Learning and Text Generation",
    long_description=long_description,
    license='Apache License Version 2.0',

    packages=[
        f"texar.{name}"
        for name in setuptools.find_packages(where='texar')
    ],
    platforms='any',

    install_requires=[
        'regex>=2018.01.10',
        'numpy',
        'requests',
        'funcsigs',
        'sentencepiece>=0.1.8',
        'mypy_extensions',
        'packaging>=19.0'
    ],
    extras_require={
        'torch': ['torch>=1.0.0'],
        'examples': [],
        'extras': ['Pillow>=3.0', 'tensorboardX>=1.8'],
    },
    package_data={
        "texar.torch": [
            "../../bin/utils/multi-bleu.perl",
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for Transformer decoder.
"""
import unittest

import torch
import torch.nn.functional as F

from texar.torch.core.layers import identity
from texar.torch.modules.decoders import decoder_helpers
from texar.torch.modules.decoders.transformer_decoders import (
    TransformerDecoder, TransformerDecoderOutput)


class TransformerDecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.TransformerDecoder`
    """

    def setUp(self):
        self._vocab_size = 101
        self._batch_size = 3
        self._max_time = 5
        self._emb_dim = 512
        self._max_decode_len = 7

        self._inputs = torch.randint(
            self._vocab_size, size=(self._batch_size, self._max_time))

        self._memory = torch.rand(
            self._batch_size, self._max_time, self._emb_dim, dtype=torch.float)
        self._memory_sequence_length = torch.randint(
            self._max_time, (self._batch_size,), dtype=torch.long)

        self._embedding = torch.rand(
            self._vocab_size, self._emb_dim, dtype=torch.float)
        self._pos_embedding = torch.rand(
            self._max_decode_len, self._emb_dim, dtype=torch.float)

        def _embedding_fn(x, y):
            x_emb = F.embedding(x, self._embedding)
            y_emb = F.embedding(y, self._pos_embedding)
            return x_emb * self._emb_dim ** 0.5 + y_emb

        self._embedding_fn = _embedding_fn

        self._output_layer = torch.rand(
            self._vocab_size, self._emb_dim, dtype=torch.float)

        self._start_tokens = torch.full(
            (self._batch_size,), 1, dtype=torch.long)
        self._end_token = 2
        self.max_decoding_length = self._max_time

        _context = [[3, 4, 5, 2, 0], [4, 3, 5, 7, 2]]
        _context_length = [4, 5]
        self._context = torch.tensor(_context)
        self._context_length = torch.tensor(_context_length)

    def test_output_layer(self):
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size, output_layer=None)
        self.assertIsInstance(decoder, TransformerDecoder)

        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            output_layer=identity)
        self.assertIsInstance(decoder, TransformerDecoder)

        tensor = torch.rand(
            self._vocab_size, self._emb_dim, dtype=torch.float)
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            output_layer=tensor)
        self.assertIsInstance(decoder, TransformerDecoder)
        self.assertEqual(decoder.vocab_size, self._vocab_size)

    def test_decode_train(self):
        """Tests train_greedy
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size, output_layer=self._output_layer)
        decoder.train()
        # 6 blocks
        # -self multihead_attention: 4 dense without bias + 2 layer norm vars
        # -encdec multihead_attention: 4 dense without bias + 2 layer norm vars
        # -poswise_network: Dense with bias, Dense with bias + 2 layer norm vars
        # 2 layer norm vars
        outputs = decoder(memory=self._memory,
                          memory_sequence_length=self._memory_sequence_length,
                          memory_attention_bias=None,
                          inputs=self._inputs,
                          decoding_strategy='train_greedy')
        # print(decoder)
        # for name, _ in decoder.named_parameters():
        #     print(name)
        self.assertEqual(len(decoder.trainable_variables), 110)
        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_decode_infer_greedy(self):
        """Tests train_greedy
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size, output_layer=self._output_layer)
        decoder.eval()
        helper = decoder_helpers.GreedyEmbeddingHelper(
            self._start_tokens, self._end_token)

        outputs, length = decoder(
            memory=self._memory,
            memory_sequence_length=self._memory_sequence_length,
            memory_attention_bias=None,
            inputs=None,
            helper=helper,
            max_decoding_length=self._max_decode_len)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_infer_greedy_with_context_without_memory(self):
        """Tests train_greedy with context
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size,
            output_layer=self._output_layer)
        decoder.eval()
        outputs, length = decoder(
            memory=None,
            memory_sequence_length=None,
            memory_attention_bias=None,
            inputs=None,
            decoding_strategy='infer_greedy',
            context=self._context,
            context_sequence_length=self._context_length,
            end_token=self._end_token,
            max_decoding_length=self._max_decode_len)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_decode_infer_sample(self):
        """Tests infer_sample
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size,
            output_layer=self._output_layer)
        decoder.eval()
        helper = decoder_helpers.SampleEmbeddingHelper(
            self._start_tokens, self._end_token)

        outputs, length = decoder(
            memory=self._memory,
            memory_sequence_length=self._memory_sequence_length,
            memory_attention_bias=None,
            inputs=None,
            helper=helper,
            max_decoding_length=self._max_decode_len)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_beam_search(self):
        """Tests beam_search
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size,
            output_layer=self._output_layer)
        decoder.eval()
        beam_width = 5
        outputs = decoder(
            memory=self._memory,
            memory_sequence_length=self._memory_sequence_length,
            memory_attention_bias=None,
            inputs=None,
            beam_width=beam_width,
            start_tokens=self._start_tokens,
            end_token=self._end_token,
            max_decoding_length=self._max_decode_len)

        self.assertEqual(outputs['log_prob'].size(),
                         (self._batch_size, beam_width))
        self.assertEqual(outputs['sample_id'].size(0), self._batch_size)
        self.assertLessEqual(outputs['sample_id'].size(2), self._max_decode_len)
        self.assertEqual(outputs['sample_id'].size(2), beam_width)

    def test_greedy_embedding_helper(self):
        """Tests with tf.contrib.seq2seq.GreedyEmbeddingHelper
        """
        decoder = TransformerDecoder(
            token_pos_embedder=self._embedding_fn,
            vocab_size=self._vocab_size,
            output_layer=self._output_layer)
        decoder.eval()
        helper = decoder_helpers.GreedyEmbeddingHelper(
            self._start_tokens, self._end_token)
        outputs, length = decoder(
            memory=self._memory,
            memory_sequence_length=self._memory_sequence_length,
            memory_attention_bias=None,
            helper=helper,
            max_decoding_length=self._max_decode_len)

        self.assertIsInstance(outputs, TransformerDecoderOutput)


if __name__ == "__main__":
    unittest.main()



# Apply from Tensorflow:
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/nest.py

"""This module can perform operations on nested structures. A nested
structure is a Python sequence, tuple (including `namedtuple`), or
dict that can contain further sequences, tuples, and dicts.
"""

import collections
from typing import TypeVar, Mapping, Union, Tuple, List, Any
import torch


TypeArg = TypeVar('TypeArg')  # type argument
NestedStructure = Any


def is_sequence(seq: Any) -> bool:
    r"""If a instance is sequance(list, tuple, excluding torch.Size),
    return True, else False.
    Args:
        seq: instance to be checked.
    Returns:
        bool, True if the input instance is sequence, otherwise False.
    """
    if isinstance(seq, torch.Size):
        return False
    return isinstance(seq, (list, tuple))


def flatten(structure: NestedStructure) -> List[Any]:
    r"""Returns a flat list from a given nested structure.
    If nest is not a sequence, tuple, or dict, then returns a single-element
    `list:[nest]`.
    In the case of dict instances, the sequence consists of the values,
    sorted by key to ensure deterministic behavior. This is true also for
    `OrderedDict` instances: their sequence order is ignored, the sorting order
    of keys is used instead. The same convention is followed in
    :func:`~texar.torch.utils.nest.pack_sequence_as`. This correctly repacks
    dictionaries and `OrderedDict`s after they have been flattened, and also
    allows flattening an `OrderedDict` and then repacking it back using a
    corresponding plain dict, or vice-versa. Dictionaries with non-sortable
    keys cannot be flattened. Users must not modify any collections used in
    nest while this function is running.

    Args:
        structure: an arbitrarily nested structure or a scalar object. Note,
            numpy arrays are considered scalars.

    Returns:
        A Python list, the flattened version of the input.

    Raises:
        TypeError: The nest is or contains a dict with non-sortable keys.
    """
    res: List[Any] = []
    if isinstance(structure, dict):
        structure = list(structure.values())

    if not is_sequence(structure):
        return [structure]
    else:
        for item in _yield_value(structure):
            res += flatten(item)

    return res


def pack_sequence_as(structure: NestedStructure,
                     flat_sequence: Union[List, Tuple]
                     ) -> NestedStructure:
    r"""Returns a given flattened sequence packed into a given structure.
    If ``structure`` is a scalar, ``flat_sequence`` must be a single-element
    list; in this case the return value is ``flat_sequence[0]``.
    If ``structure`` is or contains a dict instance, the keys will be sorted to
    pack the flat sequence in deterministic order. This is true also for
    `OrderedDict` instances: their sequence order is ignored, the sorting
    order of keys is used instead. The same convention is followed in
    :func:`~texar.torch.utils.nest.flatten`. This correctly repacks dictionaries
    and `OrderedDicts` after they have been flattened, and also allows
    flattening an `OrderedDict` and then repacking it back using a
    corresponding plain dict, or vice-versa. Dictionaries with non-sortable
    keys cannot be flattened.

    Args:
        structure: Nested structure, whose structure is given by nested lists,
            tuples, and dictionaries. Note: numpy arrays and strings are
            considered scalars.
        flat_sequence: flat sequence to pack.

    Returns:
        packed: ``flat_sequence`` converted to have the same recursive
        structure as ``structure``.

    Raises:
        ValueError: If ``flat_sequence`` and ``structure`` have different
            element counts.
        TypeError: ``structure`` is or contains a dict with non-sortable keys.
    """
    if not is_sequence(flat_sequence):
        raise TypeError("flat_sequence must be a sequence")
    if isinstance(structure, dict):
        structure = list(structure.values())
    if not is_sequence(structure):
        if len(flat_sequence) != 1:
            raise ValueError("Structure is a scalar"
                             "but len(flat_sequence) == %d > 1"
                             % len(flat_sequence))
        return flat_sequence[0]
    try:
        final_index, packed = _packed_nest_with_indices(structure,
                                                        flat_sequence, 0)
        if final_index < len(flat_sequence):
            raise IndexError
    except IndexError:
        flat_structure = flatten(structure)
        if len(flat_structure) != len(flat_sequence):
            raise ValueError(
                "Could not pack sequence. Structure had %d elements, but "
                "flat_sequence had %d elements.  Structure: %s,"
                "flat_sequence: %s." %
                (len(flat_structure),
                 len(flat_sequence),
                 structure,
                 flat_sequence))
    return _sequence_like(structure, packed)


def _sorted(dict_: Mapping[Any, Any]) -> List[TypeArg]:
    r"""Returns a sorted list of the dict keys, with error if keys not
    sortable.
    """
    try:
        return sorted(dict_)
    except TypeError:
        raise TypeError("nest only supports dicts with sortable keys.")


def _is_namedtuple(instance: object) -> bool:
    r"""Returns True if `instance` is a `namedtuple`.
    Args:
        instance: An instance of a Python object.
    Returns:
        True if `instance` is a `namedtuple`.
    """
    type_ = type(instance)
    base_ = type_.__bases__
    if len(base_) != 1 or base_[0] != tuple:
        return False
    field_ = getattr(type_, '_fields', None)
    if not isinstance(field_, tuple):
        return False
    return all(isinstance(n, str) for n in field_)


def _yield_value(iterable):
    r"""Yield only sorted values from `iterable`.
    Args:
        iterable: an iterable.
    Yields:
        The iterable's values, in order of sorted keys or items.
    """
    for _, value in _yield_sorted_items(iterable):
        yield value


def _yield_sorted_items(iterable):
    r"""Yield (key, value) pairs for `iterable` in a deterministic order.
    For Sequences, the key will be an int, the array index of a value.
    For Mappings, the key will be the dictionary key.
    For objects (e.g. namedtuples), the key will be the attribute name.
    In all cases, the keys will be iterated in sorted order.
    Args:
        iterable: an iterable.
    Yields:
        The iterable's (key, value) pairs, in order of sorted keys.
    """
    if isinstance(iterable, collections.abc.Mapping):
        for key in _sorted(iterable):
            yield key, iterable[key]

    elif _is_namedtuple(iterable):
        for field in iterable._fields:
            yield field, getattr(iterable, field)

    else:
        for index, item in enumerate(iterable):
            yield index, item


def _packed_nest_with_indices(structure: NestedStructure,
                              flat: NestedStructure,
                              index: int) -> Tuple[int, NestedStructure]:
    r"""Helper function for pack_sequence_as.
    Args:
        structure: Substructure (list / tuple / dict) to mimic.
        flat: Flattened values to output substructure for.
        index: Index at which to start reading from flat.
        is_seq: Function used to test if a value should be treated as a
            sequence.
    Returns:
        The tuple (new_index, child), where:
        * new_index - the updated index into `flat` having processed
                    `structure`.
        * packed - the subset of `flat` corresponding to `structure`,
                    having started at `index`, and packed into the same nested
                    format.
    Raises:
        ValueError: if `structure` contains more elements than `flat`
        (assuming indexing starts from `index`).
    """
    packed = []
    for value in _yield_value(structure):
        if is_sequence(value):
            new_index, child = _packed_nest_with_indices(value, flat, index)
            packed.append(_sequence_like(value, child))
            index = new_index
        else:
            packed.append(flat[index])
            index += 1
    return index, packed


InstanceType = Any


def _sequence_like(instance: InstanceType,
                   args: Any) -> InstanceType:
    r"""Converts the sequence `args` to the same type as `instance`.
    Args:
        instance: an instance of `tuple`, `list`, `namedtuple`, `dict`,
            `collections.OrderedDict`.
        args: elements to be converted to the `instance` type.
    Returns:
        `args` with the type of `instance`.
    """
    if isinstance(instance, collections.abc.Mapping):
        result: Mapping[Any, Any] = dict(zip(_sorted(instance), args))
        generator = ((key, result[key]) for key in instance)
        return type(instance)(generator)  # type: ignore
    elif _is_namedtuple(instance):
        return type(instance)(*args)
    else:
        # Not a namedtuple
        return type(instance)(args)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utils of XLNet Modules.
"""

import json
import os
from abc import ABC
from typing import Any, Callable, Dict, Optional, Union

import torch
from torch import nn

from texar.torch.modules.pretrained.pretrained_base import PretrainedMixin
from texar.torch.modules.pretrained.xlnet_utils import (
    PositionWiseFF, RelativeMultiheadAttention, init_weights)

__all__ = [
    "PretrainedXLNetMixin",
]

_XLNET_PATH = "https://storage.googleapis.com/xlnet/released_models/"


class PretrainedXLNetMixin(PretrainedMixin, ABC):
    r"""A mixin class to support loading pre-trained checkpoints for modules
    that implement the XLNet model.

    The XLNet model was proposed in
    `XLNet: Generalized Autoregressive Pretraining for Language Understanding`_
    by `Yang et al.` It is based on the Transformer-XL model, pre-trained on a
    large corpus using a language modeling objective that considers all
    permutations of the input sentence.

    The available XLNet models are as follows:

      * ``xlnet-based-cased``: 12-layer, 768-hidden, 12-heads. This model is
        trained on full data (different from the one in the paper).
      * ``xlnet-large-cased``: 24-layer, 1024-hidden, 16-heads.

    We provide the following XLNet classes:

      * :class:`~texar.torch.modules.XLNetEncoder` for text encoding.
      * :class:`~texar.torch.modules.XLNetDecoder` for text generation and
        decoding.
      * :class:`~texar.torch.modules.XLNetClassifier` for text classification
        and sequence tagging.
      * :class:`~texar.torch.modules.XLNetRegressor` for text regression.

    .. _`XLNet: Generalized Autoregressive Pretraining for Language Understanding`:
        http://arxiv.org/abs/1906.08237
    """
    _MODEL_NAME = "XLNet"
    _MODEL2URL = {
        'xlnet-base-cased':
            _XLNET_PATH + "cased_L-12_H-768_A-12.zip",
        'xlnet-large-cased':
            _XLNET_PATH + "cased_L-24_H-1024_A-16.zip",
    }

    def reset_parameters(self):
        self.apply(init_weights)
        if not self._hparams.untie_r:
            nn.init.normal_(self.r_w_bias, 0.0, 0.02)
            nn.init.normal_(self.r_r_bias, 0.0, 0.02)
            if self._hparams.use_segments:
                nn.init.normal_(self.r_s_bias, 0.0, 0.02)

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str) -> Dict[str, Any]:
        info = list(os.walk(cache_dir))
        root, _, files = info[0]
        config_path = None
        for file in files:
            if file.endswith('config.json'):
                config_path = os.path.join(root, file)
        if config_path is None:
            raise ValueError(f"Cannot find the config file in {cache_dir}")

        with open(config_path) as f:
            config_ckpt = json.loads(f.read())

        configs = {
            "head_dim": config_ckpt["d_head"],
            "ffn_inner_dim": config_ckpt["d_inner"],
            "hidden_dim": config_ckpt["d_model"],
            "activation": config_ckpt["ff_activation"],
            "num_heads": config_ckpt["n_head"],
            "num_layers": config_ckpt["n_layer"],
            "vocab_size": config_ckpt["n_token"],
            "untie_r": config_ckpt["untie_r"]
        }

        return configs

    def _init_from_checkpoint(self, pretrained_model_name: str,
                              cache_dir: str, **kwargs):
        # remember to call .contiguous after trans_fn
        try:
            import numpy as np
            import tensorflow as tf
        except ImportError:
            print("Loading TensorFlow models in PyTorch requires installing "
                  "TensorFlow. Please see https://www.tensorflow.org/install/ "
                  "for installation instructions.")
            raise

        ckpt = tf.train.load_checkpoint(
            os.path.join(cache_dir, 'xlnet_model.ckpt'))
        from_params: Dict[str, np.ndarray] = {
            key: ckpt.get_tensor(key)
            for key in ckpt.get_variable_to_shape_map().keys()}
        del from_params["global_step"]  # useless variable
        to_params: Dict[str, nn.Parameter] = dict(self.named_parameters())

        def get_weight(name: str) -> torch.Tensor:
            weight = from_params["model/" + name]
            del from_params["model/" + name]
            return torch.from_numpy(weight)

        TransFn = Callable[[torch.Tensor], torch.Tensor]

        def assign(param: nn.Parameter, weight: Union[str, torch.Tensor],
                   trans_fn: Optional[TransFn] = None,
                   allow_fail: bool = False):
            param_key = next(k for k, v in to_params.items() if v is param)
            # Delete regardless of whether weight exists.
            del to_params[param_key]
            if isinstance(weight, str):
                try:
                    weight = get_weight(weight)
                except KeyError:
                    if allow_fail:
                        print(f"Weight {weight} not found in checkpoint")
                        return
                    else:
                        raise
            if trans_fn is not None:
                weight = trans_fn(weight).contiguous()
            if param.size() != weight.size():
                raise ValueError(f"Expected size {param.size()}, "
                                 f"actual size {weight.size()}")
            param.data = weight

        def assign_linear(linear: nn.Linear, prefix: str):
            trans_fn = lambda p: p.view(p.size(0), -1).t()
            assign(linear.weight, prefix + "kernel", trans_fn)
            if linear.bias is not None:
                assign(linear.bias, prefix + "bias")

        def assign_layer_norm(layer_norm: nn.LayerNorm, prefix: str):
            assign(layer_norm.weight, prefix + "LayerNorm/gamma")
            assign(layer_norm.bias, prefix + "LayerNorm/beta")

        def load_xlnet_model(xlnet):
            n_layers = len(xlnet.attn_layers)
            for bias_name in ['r_r_bias', 'r_w_bias', 'r_s_bias']:
                weight = get_weight("transformer/" + bias_name)
                if xlnet.hparams.untie_r:
                    for idx in range(n_layers):
                        layer: RelativeMultiheadAttention
                        layer = xlnet.attn_layers[idx]
                        assign(getattr(layer, bias_name), weight[idx])
                else:
                    assign(getattr(xlnet, bias_name), weight)
            assign(xlnet.word_embed.weight,
                   "transformer/word_embedding/lookup_table")

            for idx in range(n_layers):
                layer: RelativeMultiheadAttention = xlnet.attn_layers[idx]
                prefix = f"transformer/layer_{idx}/rel_attn/"
                qkv_weights = [get_weight(prefix + f"{part}/kernel")
                               for part in "qkv"]
                assign(layer.head_projection.weight,
                       torch.cat([
                           p.view(p.size(0), -1) for p in qkv_weights
                       ], dim=1).t())
                assign_linear(layer.pos_projection, prefix + "r/")
                assign(layer.output_projection.weight,  # DO NOT TRANSPOSE!!!!
                       prefix + "o/kernel", lambda p: p.view(p.size(0), -1))
                assign_layer_norm(layer.layer_norm, prefix)

            for idx in range(n_layers):
                layer: PositionWiseFF = xlnet.ff_layers[idx]
                prefix = f"transformer/layer_{idx}/ff/"
                for linear_idx in range(1, 2 + 1):
                    linear_prefix = f"{prefix}layer_{linear_idx}/"
                    linear_layer: nn.Linear = getattr(
                        layer, f"linear{linear_idx}")
                    assign_linear(linear_layer, linear_prefix)
                assign_layer_norm(layer.layer_norm, prefix)

            seg_embeds = [
                p.squeeze(0)
                for p in torch.chunk(
                    get_weight("transformer/seg_embed"), n_layers, dim=0)]
            for idx in range(n_layers):
                assign(xlnet.attn_layers[idx].segment_embed, seg_embeds[idx])

            if hasattr(xlnet, 'mask_emb') and hasattr(xlnet, 'lm_bias'):
                assign(xlnet.mask_emb, "transformer/mask_emb/mask_emb")
                assign(xlnet.lm_bias, "lm_loss/bias")

        load_xlnet_model(self)

        if len(from_params) > 0:
            print(f"WARNING: Certain weights from checkpoint are not loaded: "
                  f"{list(from_params.keys())}")

        filtered_to_params = [k for k in to_params if k.startswith("xlnet")]
        if len(filtered_to_params) > 0:
            print(f"WARNING: Certain parameters are not initialized: "
                  f"{list(filtered_to_params)}")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pre-trained RoBERTa tokenizer.
"""

from typing import Any, Dict, List, Optional, Tuple

from texar.torch.data.tokenizers.gpt2_tokenizer import GPT2Tokenizer
from texar.torch.utils.utils import truncate_seq_pair

__all__ = [
    'RoBERTaTokenizer',
]

_ROBERTA_PATH = "https://s3.amazonaws.com/models.huggingface.co/bert/"


class RoBERTaTokenizer(GPT2Tokenizer):
    r"""Pre-trained RoBERTa Tokenizer.

    Args:
        pretrained_model_name (optional): a `str`, the name of
            pre-trained model (e.g., `roberta-base`). Please refer to
            :class:`~texar.torch.modules.PretrainedRoBERTaMixin` for
            all supported models.
            If None, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    _MODEL2URL = {
        'roberta-base': [
            _ROBERTA_PATH + 'roberta-base-vocab.json',
            _ROBERTA_PATH + 'roberta-base-merges.txt',
        ],
        'roberta-large': [
            _ROBERTA_PATH + 'roberta-large-vocab.json',
            _ROBERTA_PATH + 'roberta-large-merges.txt',
        ],
    }
    _MAX_INPUT_SIZE = {
        'roberta-base': 512,
        'roberta-large': 512,
    }
    _VOCAB_FILE_MAP = {
        'vocab_file': {
            'roberta-base': 'roberta-base-vocab.json',
            'roberta-large': 'roberta-large-vocab.json',
        },
        'merges_file': {
            'roberta-base': 'roberta-base-merges.txt',
            'roberta-large': 'roberta-large-merges.txt',
        },
    }

    def encode_text(self,  # type: ignore
                    text_a: str,
                    text_b: Optional[str] = None,
                    max_seq_length: Optional[int] = None) -> \
            Tuple[List[int], List[int]]:
        r"""Adds special tokens to a sequence or sequence pair and computes the
        corresponding input mask for RoBERTa specific tasks.
        The sequence will be truncated if its length is larger than
        ``max_seq_length``.

        A RoBERTa sequence has the following format:
        `[cls_token]` X `[sep_token]`

        A RoBERTa sequence pair has the following format:
        `[cls_token]` A `[spe_token]` `[sep_token]` B `[sep_token]`

        Args:
            text_a: The first input text.
            text_b: The second input text.
            max_seq_length: Maximum sequence length.

        Returns:
            A tuple of `(input_ids, segment_ids, input_mask)`, where

            - ``input_ids``: A list of input token ids with added
              special token ids.
            - ``input_mask``: A list of mask ids. The mask has 1 for real
              tokens and 0 for padding tokens. Only real tokens are
              attended to.
        """
        if max_seq_length is None:
            max_seq_length = self.max_len

        cls_token_id = self._map_token_to_id(self.cls_token)
        sep_token_id = self._map_token_to_id(self.sep_token)

        token_ids_a = self.map_text_to_id(text_a)
        assert isinstance(token_ids_a, list)

        token_ids_b = None
        if text_b:
            token_ids_b = self.map_text_to_id(text_b)

        if token_ids_b:
            assert isinstance(token_ids_b, list)
            # Modifies `token_ids_a` and `token_ids_b` in place so that the
            # total length is less than the specified length.
            # Account for <s>, </s>, </s>, </s> with "- 4"
            truncate_seq_pair(token_ids_a, token_ids_b, max_seq_length - 4)

            input_ids = ([cls_token_id] + token_ids_a + [sep_token_id] +
                         [sep_token_id] + token_ids_b + [sep_token_id])
        else:
            # Account for <s> and </s> with "- 2"
            token_ids_a = token_ids_a[:max_seq_length - 2]

            input_ids = [cls_token_id] + token_ids_a + [sep_token_id]

        input_mask = [1] * len(input_ids)

        # Zero-pad up to the maximum sequence length.
        input_ids = input_ids + [0] * (max_seq_length - len(input_ids))
        input_mask = input_mask + [0] * (max_seq_length - len(input_mask))

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length

        return input_ids, input_mask

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * The tokenizer is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the tokenizer is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the tokenizer is defined by the
          configurations in `hparams`.

        .. code-block:: python

            {
                "pretrained_model_name": "roberta-base",
                "vocab_file": None,
                "merges_file": None,
                "max_len": 512,
                "bos_token": "<s>",
                "eos_token": "</s>",
                "sep_token": "</s>",
                "cls_token": "</s>",
                "unk_token": "<unk>",
                "pad_token": "<pad>",
                "mask_token": "<mask>",
                "errors": "replace",
                "name": "roberta_tokenizer",
            }

        Here:

        `"pretrained_model_name"`: str or None
            The name of the pre-trained RoBERTa model.

        `"vocab_file"`: str or None
            The path to a vocabulary json file mapping tokens to ids.

        `"merges_file"`: str or None
            The path to a merges file.

        `"max_len"`: int
            The maximum sequence length that this model might ever be used with.

        `"bos_token"`: str
            Beginning of sentence token.

        `"eos_token"`: str
            End of sentence token.

        `"sep_token"`: str
            Separation token.

        `"cls_token"`: str
            Classification token.

        `"unk_token"`: str
            Unknown token.

        `"pad_token"`: str
            Padding token.

        `"mask_token"`: str
            Masking token.

        `"errors"`: str
            Response when decoding fails. The possible values are
            `ignore`, `replace`, and `strict`.

        `"name"`: str
            Name of the tokenizer.
        """
        return {
            'pretrained_model_name': 'roberta-base',
            'vocab_file': None,
            'merges_file': None,
            'max_len': 512,
            'bos_token': '<s>',
            'eos_token': '</s>',
            'sep_token': '</s>',
            'cls_token': '</s>',
            'unk_token': '<unk>',
            'pad_token': '<pad>',
            'mask_token': '<mask>',
            'errors': 'replace',
            'name': 'roberta_tokenizer',
            '@no_typecheck': ['pretrained_model_name'],
        }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str):
        r"""Returns the configuration of the pre-trained RoBERTa tokenizer."""
        return {
            'vocab_file': None,
            'merges_file': None,
            'max_len': 512,
            'bos_token': '<s>',
            'eos_token': '</s>',
            'sep_token': '</s>',
            'cls_token': '<s>',
            'unk_token': '<unk>',
            'pad_token': '<pad>',
            'mask_token': '<mask>',
            'errors': 'replace',
        }



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for GPT2 decoder.
"""
import unittest

import torch

from texar.torch.modules.decoders import decoder_helpers
from texar.torch.modules.decoders.gpt2_decoder import GPT2Decoder
from texar.torch.modules.decoders.transformer_decoders import \
    TransformerDecoderOutput
from texar.torch.utils.test import pretrained_test


class GPT2DecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.GPT2Decoder`
    """

    def setUp(self) -> None:
        # Use small prime numbers to speedup tests.
        self.batch_size = 2
        self.max_length = 3
        self.beam_width = 5
        self.inputs = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_hparams(self):
        r"""Tests the priority of the decoder arch parameters.
        """
        # case 1: set "pretrained_mode_name" by constructor argument
        hparams = {
            "pretrained_model_name": "gpt2-medium",
        }
        decoder = GPT2Decoder(pretrained_model_name="gpt2-small",
                              hparams=hparams)
        self.assertEqual(decoder.hparams.decoder.num_blocks, 12)
        _ = decoder(self.inputs)

        # case 2: set "pretrained_mode_name" by hparams
        hparams = {
            "pretrained_model_name": "gpt2-small",
            "decoder": {
                "num_blocks": 6,
            },
        }
        decoder = GPT2Decoder(hparams=hparams)
        self.assertEqual(decoder.hparams.decoder.num_blocks, 12)
        _ = decoder(self.inputs)

        # case 3: set to None in both hparams and constructor argument
        hparams = {
            "pretrained_model_name": None,
            "decoder": {
                "num_blocks": 6,
            },
        }
        decoder = GPT2Decoder(hparams=hparams)
        self.assertEqual(decoder.hparams.decoder.num_blocks, 6)
        _ = decoder(self.inputs)

        # case 4: using default hparams
        decoder = GPT2Decoder()
        self.assertEqual(decoder.hparams.decoder.num_blocks, 12)
        _ = decoder(self.inputs)

    @pretrained_test
    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        def get_variable_num(n_layers: int) -> int:
            return 1 + 1 + n_layers * 26 + 2

        # case 1: GPT2 small
        decoder = GPT2Decoder()
        self.assertEqual(len(decoder.trainable_variables), get_variable_num(12))
        _ = decoder(self.inputs)

        # case 2: GPT2 medium
        hparams = {
            "pretrained_model_name": "gpt2-medium",
        }
        decoder = GPT2Decoder(hparams=hparams)
        self.assertEqual(len(decoder.trainable_variables), get_variable_num(24))
        _ = decoder(self.inputs)

        # case 2: GPT2 large
        hparams = {
            "pretrained_model_name": "gpt2-large",
        }
        decoder = GPT2Decoder(hparams=hparams)
        self.assertEqual(len(decoder.trainable_variables), get_variable_num(36))
        _ = decoder(self.inputs)

        # case 3: self-designed GPT2
        hparams = {
            "pretrained_model_name": None,
            "decoder": {
                "num_blocks": 6,
            },
        }
        decoder = GPT2Decoder(hparams=hparams)
        self.assertEqual(len(decoder.trainable_variables), get_variable_num(6))
        _ = decoder(self.inputs)

    def test_decode_train(self):
        r"""Tests train_greedy.
        """
        hparams = {
            "pretrained_model_name": None
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.train()

        inputs = torch.randint(50257, (self.batch_size, self.max_length))
        outputs = decoder(inputs)

        self.assertEqual(outputs.logits.shape,
                         torch.Size([self.batch_size, self.max_length, 50257]))
        self.assertEqual(outputs.sample_id.shape,
                         torch.Size([self.batch_size, self.max_length]))

    def test_decode_infer_greedy(self):
        r"""Tests train_greedy
        """
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.eval()

        start_tokens = torch.full((self.batch_size,), 1, dtype=torch.int64)
        end_token = 2

        helper = decoder_helpers.GreedyEmbeddingHelper(start_tokens, end_token)

        outputs, length = decoder(
            helper=helper, max_decoding_length=self.max_length)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_decode_infer_sample(self):
        r"""Tests infer_sample
        """
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.eval()

        start_tokens = torch.full((self.batch_size,), 1, dtype=torch.int64)
        end_token = 2

        helper = decoder_helpers.SampleEmbeddingHelper(start_tokens, end_token)

        outputs, length = decoder(
            helper=helper, max_decoding_length=self.max_length)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_beam_search(self):
        r"""Tests beam_search
        """
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.eval()

        start_tokens = torch.full((self.batch_size,), 1, dtype=torch.int64)
        end_token = 2

        outputs = decoder(
            start_tokens=start_tokens, beam_width=self.beam_width,
            end_token=end_token, max_decoding_length=self.max_length)

        self.assertEqual(
            outputs['log_prob'].shape,
            torch.Size([self.batch_size, self.beam_width]))
        self.assertEqual(outputs['sample_id'].shape[0], self.batch_size)
        self.assertLessEqual(outputs['sample_id'].shape[1], self.max_length)
        self.assertEqual(outputs['sample_id'].shape[2], self.beam_width)

    def test_greedy_embedding_helper(self):
        r"""Tests with tf.contrib.seq2seq.GreedyEmbeddingHelper
        """
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.eval()

        start_tokens = torch.full((self.batch_size,), 1, dtype=torch.int64)
        end_token = 2

        helper = decoder_helpers.GreedyEmbeddingHelper(start_tokens, end_token)

        outputs, length = decoder(
            helper=helper, max_decoding_length=self.max_length)

        self.assertIsInstance(outputs, TransformerDecoderOutput)

    def test_topk_embedding_helper(self):
        r"""Tests TopKSampleEmbeddingHelper
        """
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = GPT2Decoder(hparams=hparams)
        decoder.eval()

        start_tokens = torch.full((self.batch_size,), 1, dtype=torch.int64)
        end_token = 2

        helper = decoder_helpers.TopKSampleEmbeddingHelper(
            start_tokens=start_tokens, end_token=end_token,
            top_k=40, softmax_temperature=0.7)

        outputs, length = decoder(
            helper=helper, max_decoding_length=self.max_length)

        self.assertIsInstance(outputs, TransformerDecoderOutput)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utils for T5 Modules
"""

import copy
import os
from abc import ABC
from typing import Any, Dict, List, Set

import numpy as np
import torch

from texar.torch.modules.pretrained.pretrained_base import PretrainedMixin
from texar.torch.modules.pretrained.t5_utils import read_t5_gin_config_file

__all__ = [
    "PretrainedT5Mixin"
]

_T5_PATH = "https://storage.googleapis.com/t5-data/pretrained_models/"
_T5_VOCAB_PATH = "https://storage.googleapis.com/t5-data/vocabs/cc_all.32000/"
_CHECKPOINT_FILES_GEN_MAP = {  # stores a tuple of model_id and number of
                               # partitions
    'small': (1000000, 16),
    'base': (999900, 16),
    'large': (1000700, 8),
    'B': (1000000, 64)
}


def _generate_t5_file_list(ckpt_tuple: tuple) -> List[str]:
    """ Helper function to generate file list given a tuple of model_id and
    partition size.

    Args:
        ckpt_tuple: A tuple of model_id and number of partitions

    """
    ckpt_id = ckpt_tuple[0]
    ckpt_parts = ckpt_tuple[1]
    return [
        'checkpoint',
        *[f'model.ckpt-{ckpt_id}.data-{idx:05d}-of-{ckpt_parts:05d}'
          for idx in range(ckpt_parts)],
        f'model.ckpt-{ckpt_id}.index',
        f'model.ckpt-{ckpt_id}.meta',
        'operative_config.gin']


class PretrainedT5Mixin(PretrainedMixin, ABC):
    r"""A mixin class to support loading pre-trained checkpoints for modules
        that implement the T5 model.

    The T5 model treats multiple NLP tasks in a similar manner by encoding the
    different tasks as text directives in the input stream. This enables a
    single model to be trained supervised on a wide variety of NLP tasks.

    The T5 model examines factors relevant for leveraging transfer learning
    at scale from pure unsupervised pre-training to supervised tasks. It is
    discussed in much detail in `Exploring the Limits of Transfer Learning
    with a Unified Text-to-Text Transformer` from Google.

    The available T5 models are as follows:

      * ``T5-Small``: Small version of T5, 60 million parameters.
      * ``T5-Base``: Base-line version of T5, 220 million parameters.
      * ``T5-Large``: Large Version of T5, 770 million parameters.
      * ``T5-3B``: A version of T5 with 3 billion parameters.
      * ``T5-11B``: A version of T5 with 11 billion parameters.

    We provide the following classes:

      * :class:`~texar.torch.modules.T5Encoder` for loading weights for the
        encoder stack.
      * :class:`~texar.torch.modules.T5Decoder` for loading weights for the
        decoding stack.
      * :class:`~texar.torch.modules.T5EncoderDecoder` as a raw pre-trained
        model.
    """
    _MODEL_NAME = "T5"

    _MODEL2URL = {
        'T5-Small': [_T5_PATH + f"small/{file}"
                     for file in _generate_t5_file_list(
                      _CHECKPOINT_FILES_GEN_MAP['small'])] +
                    [_T5_VOCAB_PATH + 'sentencepiece.model'],
        'T5-Base': [_T5_PATH + f"base/{file}"
                    for file in _generate_t5_file_list(
                     _CHECKPOINT_FILES_GEN_MAP['base'])] +
                   [_T5_VOCAB_PATH + 'sentencepiece.model'],
        'T5-Large': [_T5_PATH + f"large/{file}"
                     for file in _generate_t5_file_list(
                      _CHECKPOINT_FILES_GEN_MAP['large'])] +
                    [_T5_VOCAB_PATH + 'sentencepiece.model'],
        'T5-3B': [_T5_PATH + f"3B/{file}"
                  for file in _generate_t5_file_list(
                   _CHECKPOINT_FILES_GEN_MAP['B'])] + [_T5_VOCAB_PATH +
                                                       'sentencepiece.model'],
        'T5-11B': [_T5_PATH + f"11B/{file}"
                   for file in _generate_t5_file_list(
                    _CHECKPOINT_FILES_GEN_MAP['B'])] +
                  [_T5_VOCAB_PATH + 'sentencepiece.model']
    }

    _MODEL2CKPT = {
        'T5-Small': 'model.ckpt-1000000',
        'T5-Base': 'model.ckpt-999900',
        'T5-Large': 'model.ckpt-1000700',
        'T5-3B': 'model.ckpt-1000000',
        'T5-11B': 'model.ckpt-1000000'
    }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str) -> Dict[str, Any]:
        info = list(os.walk(cache_dir))
        root, _, files = info[0]
        config_path = None
        for file in files:
            if file.endswith('operative_config.gin'):
                config_path = os.path.join(root, file)
        if config_path is None:
            raise ValueError(f"Cannot find the config file in {cache_dir}")

        gin_config = read_t5_gin_config_file(config_path)

        hidden_dim = gin_config['d_model']
        vocab_size = 32128
        eps = 1e-6
        embedding_dropout = gin_config['dropout_rate']
        num_blocks = gin_config['num_layers']
        num_heads = gin_config['num_heads']
        num_units = gin_config['d_kv'] * num_heads
        dropout_rate = gin_config['dropout_rate']
        residual_dropout = gin_config['dropout_rate']
        intermediate_size = gin_config['d_ff']
        rel_attn_num_buckets = 32
        use_bias = False

        configs = {
            'hidden_size': hidden_dim,
            'embed': {
                'name': 'word_embeddings',
                'dim': hidden_dim
            },
            'vocab_size': vocab_size,

            'encoder': {
                'name': 'encoder',
                'embedding_dropout': embedding_dropout,
                'num_blocks': num_blocks,
                'multihead_attention': {
                    'use_bias': use_bias,
                    'num_units': num_units,
                    'num_heads': num_heads,
                    'output_dim': hidden_dim,
                    'dropout_rate': dropout_rate,
                    'name': 'self',
                    'is_decoder': False,
                    'relative_attention_num_buckets': rel_attn_num_buckets
                },
                'eps': eps,
                'residual_dropout': residual_dropout,
                'dim': hidden_dim,
                'poswise_feedforward': {
                    "layers": [{
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': hidden_dim,
                            'out_features': intermediate_size,
                            'bias': use_bias,
                        }
                    }, {
                        'type': "ReLU"
                    }, {
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': intermediate_size,
                            'out_features': hidden_dim,
                            'bias': use_bias,
                        }
                    }],
                },
            },
            'decoder': {
                'name': 'decoder',
                'embedding_dropout': embedding_dropout,
                'num_blocks': num_blocks,
                'multihead_attention': {
                    'use_bias': use_bias,
                    'num_units': num_units,
                    'num_heads': num_heads,
                    'output_dim': hidden_dim,
                    'dropout_rate': dropout_rate,
                    'name': 'self',
                    'is_decoder': True,
                    'relative_attention_num_buckets': rel_attn_num_buckets
                },
                'eps': eps,
                'residual_dropout': residual_dropout,
                'dim': hidden_dim,
                'poswise_feedforward': {
                    "layers": [{
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': hidden_dim,
                            'out_features': intermediate_size,
                            'bias': use_bias,
                        }
                    }, {
                        'type': "ReLU"
                    }, {
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': intermediate_size,
                            'out_features': hidden_dim,
                            'bias': use_bias,
                        }
                    }],
                },
            }

        }

        return configs

    def assign(self, from_array, to_param, transpose=False):
        pointer = self._name_to_variable(to_param)
        if transpose:
            from_array = np.transpose(from_array)
        assert pointer.shape == from_array.shape
        pointer.data = torch.from_numpy(from_array.astype(np.float32))

    def _init_from_checkpoint(self, pretrained_model_name: str,
                              cache_dir: str, **kwargs):
        try:
            import tensorflow as tf
        except ImportError:
            print("Loading TensorFlow models in PyTorch requires installing "
                  "TensorFlow. Please see https://www.tensorflow.org/install/ "
                  "for installation instructions.")
            raise

        tf_path = os.path.abspath(os.path.join(
            cache_dir,
            self._MODEL2CKPT[pretrained_model_name]))

        # Load weights from TF model
        init_vars = tf.train.list_variables(tf_path)

        to_params: Set[str] = {x[0] for x in self.named_parameters()}
        to_params.remove(  # Not used as duplicate weights stored
            'decoder.enc_dec_attns.0.relative_attention_bias.weight')

        tfnames, arrays = [], []
        for name, _ in init_vars:
            array = tf.train.load_variable(tf_path, name)
            tfnames.append(name)
            arrays.append(array.squeeze())

        from_params = set(copy.deepcopy(tfnames))
        global_tensor_map = {
            'shared/embedding': 'word_embedder._embedding'
        }
        self_attention_map = {
            'SelfAttention/k': '{}.self_attns.{}.K_dense.weight',
            'SelfAttention/o': '{}.self_attns.{}.O_dense.weight',
            'SelfAttention/q': '{}.self_attns.{}.Q_dense.weight',
            'SelfAttention/v': '{}.self_attns.{}.V_dense.weight',
            'SelfAttention/relative_attention_bias':
                '{}.self_attns.{}.relative_attention_bias.weight',
            'layer_norm/scale': '{}.self_attn_layer_norm.{}.w'
        }

        enc_dec_attention_map = {
            'EncDecAttention/k': '{}.enc_dec_attns.{}.K_dense.weight',
            'EncDecAttention/o': '{}.enc_dec_attns.{}.O_dense.weight',
            'EncDecAttention/q': '{}.enc_dec_attns.{}.Q_dense.weight',
            'EncDecAttention/v': '{}.enc_dec_attns.{}.V_dense.weight',
            'layer_norm/scale': '{}.end_dec_attn_layer_norm.{}.w'
        }

        drd_map = {
            'DenseReluDense/wi/kernel':
                '{}.poswise_networks.{}._layers.0.weight',
            'DenseReluDense/wo/kernel':
                '{}.poswise_networks.{}._layers.2.weight',
            'layer_norm/scale': '{}.poswise_layer_norm.{}.w'
        }

        component_map = {  # For encoder/decoder level
            'final_layer_norm/scale': '{}.final_layer_norm.w'
        }

        block_map = {
            'encoder0': self_attention_map,
            'encoder1': drd_map,
            'decoder0': self_attention_map,
            'decoder1': enc_dec_attention_map,
            'decoder2': drd_map,
        }

        idx = 0

        # Initialize this param separately
        special_param_name = \
            'decoder.enc_dec_attns.0.relative_attention_bias.weight'
        rab_pointer = self._name_to_variable(special_param_name)
        rab_pointer.data.normal_(mean=0.0,
                                 std=(self._hparams.hidden_size) ** -0.5)

        for name, array in zip(tfnames, arrays):
            if name.startswith('cls') or name == 'global_step' or \
                    name.endswith('adam_m') or name.endswith('adam_v')\
                    or '_slot_' in name:
                # ignore those variables begin with cls
                # ignore 'global_step' variable
                # ignore optimizer state variable
                # ignore slot
                from_params.remove(name)
                continue

            if name in global_tensor_map:
                v_name = global_tensor_map[name]
                self.assign(array, v_name)
                idx += 1
                from_params.remove(name)
                to_params.remove(v_name)
            else:
                # e.g. decoder/block_000/layer_000/SelfAttention/k
                tmp_name = name.split('/')
                submodule = tmp_name[0]  # encoder/decoder
                if len(tmp_name) > 3:  # has block level data
                    block_num = int(tmp_name[1][6:])
                    layer_num = str(int(tmp_name[2][6:]))
                    sublayer_name = "/".join(tmp_name[3:])
                    # Block-wise params
                    map_ = block_map[submodule + layer_num]
                    if sublayer_name in map_:
                        v_name = map_[sublayer_name].format(submodule,
                                                            block_num)
                        self.assign(array, v_name, True)
                        idx += 1
                        from_params.remove(name)
                        to_params.remove(v_name)
                else:
                    # e.g. decoder/final_layer_norm/scale
                    sublayer_name = "/".join(tmp_name[1:])
                    if sublayer_name in component_map:
                        v_name = component_map[sublayer_name].format(submodule)
                        self.assign(array, v_name)
                        idx += 1
                        from_params.remove(name)
                        to_params.remove(v_name)

        if len(from_params) > 0:
            print(f"WARNING: Certain weights from checkpoint are not loaded: "
                  f"{list(from_params)}")

        if len(to_params) > 0:
            print(f"WARNING: Certain parameters are not initialized: "
                  f"{list(to_params)}")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for encoder_decoders.
"""
from abc import ABC
from typing import Any, Dict

from texar.torch.module_base import ModuleBase

__all__ = [
    'EncoderDecoderBase',
]


class EncoderDecoderBase(ModuleBase, ABC):
    r"""Base class inherited by all encoderdecoder classes.
    """

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.
        """
        return {
            'name': 'encoderdecoder'
        }



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for classification related operations.
"""
import functools
import unittest

import numpy as np

from texar.torch.run.executor import make_deterministic
from texar.torch.run.metric.classification import *
from texar.torch.utils.test import external_library_test


@external_library_test("sklearn")
class ClassificationMetricTest(unittest.TestCase):
    def setUp(self) -> None:
        make_deterministic(0)
        self.n_classes = 10
        self.n_examples = 300
        self.batch_size = 2
        self.labels = np.random.randint(self.n_classes, size=self.n_examples)
        self.guesses = np.random.randint(self.n_classes, size=self.n_examples)
        self.binary_labels = np.random.randint(2, size=self.n_examples)
        self.binary_guesses = np.random.randint(2, size=self.n_examples)

    def _test_metric(self, metric, reference_fn, binary=False):
        guesses = self.binary_guesses if binary else self.guesses
        labels = self.binary_labels if binary else self.labels
        for idx in range(0, self.n_examples, self.batch_size):
            end_idx = idx + self.batch_size
            metric.add(guesses[idx:end_idx], labels[idx:end_idx])
            value = metric.value()
            answer = reference_fn(labels[:end_idx], guesses[:end_idx])
            self.assertAlmostEqual(value, answer)

    def test_accuracy(self):
        from sklearn.metrics import accuracy_score
        metric = Accuracy(pred_name="")
        self._test_metric(metric, accuracy_score)

    def test_precision(self):
        from sklearn.metrics import precision_score
        for mode in Precision._valid_modes:
            metric = Precision(mode=mode, pos_label=1, pred_name="")
            self._test_metric(
                metric, functools.partial(precision_score, average=mode),
                binary=(mode == 'binary'))

    def test_recall(self):
        from sklearn.metrics import recall_score
        for mode in Recall._valid_modes:
            metric = Recall(mode=mode, pos_label=1, pred_name="")
            self._test_metric(
                metric, functools.partial(recall_score, average=mode),
                binary=(mode == 'binary'))

    def test_f1(self):
        from sklearn.metrics import f1_score
        for mode in F1._valid_modes:
            metric = F1(mode=mode, pos_label=1, pred_name="")
            self._test_metric(
                metric, functools.partial(f1_score, average=mode),
                binary=(mode == 'binary'))



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for RNN decoders.
"""

from typing import Optional, Tuple, TypeVar

import torch
from torch import nn

from texar.torch.core import layers
from texar.torch.core.cell_wrappers import RNNCellBase
from texar.torch.modules.decoders import decoder_helpers
from texar.torch.modules.decoders.decoder_base import (
    DecoderBase, TokenEmbedder, TokenPosEmbedder, _make_output_layer)
from texar.torch.modules.decoders.decoder_helpers import Helper
from texar.torch.utils import utils

__all__ = [
    'RNNDecoderBase',
]

State = TypeVar('State')
Output = TypeVar('Output')  # output type can be of any nested structure


class RNNDecoderBase(DecoderBase[State, Output]):
    r"""Base class inherited by all RNN decoder classes.
    See :class:`~texar.torch.modules.BasicRNNDecoder` for the arguments.

    See :meth:`forward` for the inputs and outputs of RNN decoders in general.
    """

    def __init__(self,
                 input_size: int,
                 vocab_size: int,
                 token_embedder: Optional[TokenEmbedder] = None,
                 token_pos_embedder: Optional[TokenPosEmbedder] = None,
                 cell: Optional[RNNCellBase] = None,
                 output_layer: Optional[nn.Module] = None,
                 input_time_major: bool = False,
                 output_time_major: bool = False,
                 hparams=None):
        super().__init__(token_embedder, token_pos_embedder,
                         input_time_major, output_time_major, hparams=hparams)

        self._input_size = input_size
        self._vocab_size = vocab_size

        # Make RNN cell
        self._cell = cell or layers.get_rnn_cell(
            input_size, self._hparams.rnn_cell)
        self._beam_search_cell = None

        # Make the output layer
        self._output_layer, _ = _make_output_layer(
            output_layer, self._vocab_size, self._cell.hidden_size,
            self._hparams.output_layer_bias)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        The hyperparameters are the same as in
        :meth:`~texar.torch.modules.BasicRNNDecoder.default_hparams` of
        :class:`~texar.torch.modules.BasicRNNDecoder`, except that the default
        ``"name"`` here is ``"rnn_decoder"``.
        """
        return {
            'rnn_cell': layers.default_rnn_cell_hparams(),
            'helper_train': decoder_helpers.default_helper_train_hparams(),
            'helper_infer': decoder_helpers.default_helper_infer_hparams(),
            'max_decoding_length_train': None,
            'max_decoding_length_infer': None,
            'name': 'rnn_decoder',
            "output_layer_bias": True,
        }

    def forward(self,  # type: ignore
                inputs: Optional[torch.Tensor] = None,
                sequence_length: Optional[torch.LongTensor] = None,
                initial_state: Optional[State] = None,
                helper: Optional[Helper] = None,
                max_decoding_length: Optional[int] = None,
                impute_finished: bool = False,
                infer_mode: Optional[bool] = None, **kwargs) \
            -> Tuple[Output, Optional[State], torch.LongTensor]:
        r"""Performs decoding. This is a shared interface for both
        :class:`~texar.torch.modules.BasicRNNDecoder` and
        :class:`~texar.torch.modules.AttentionRNNDecoder`.

        Implementation calls :meth:`initialize` once and :meth:`step`
        repeatedly on the decoder object. Please refer to
        `tf.contrib.seq2seq.dynamic_decode
        <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/dynamic_decode>`_.

        See Also:
            Arguments of :meth:`create_helper`, for arguments like
            :attr:`decoding_strategy`.

        Args:
            inputs (optional): Input tensors for teacher forcing decoding.
                Used when :attr:`decoding_strategy` is set to
                ``"train_greedy"``, or when `hparams`-configured helper is used.

                The :attr:`inputs` is a :tensor:`LongTensor` used as index to
                look up embeddings and feed in the decoder. For example, if
                :attr:`embedder` is an instance of
                :class:`~texar.torch.modules.WordEmbedder`, then :attr:`inputs`
                is usually a 2D int Tensor `[batch_size, max_time]` (or
                `[max_time, batch_size]` if `input_time_major` == `True`)
                containing the token indexes.
            sequence_length (optional): A 1D int Tensor containing the
                sequence length of :attr:`inputs`.
                Used when `decoding_strategy="train_greedy"` or
                `hparams`-configured helper is used.
            initial_state (optional): Initial state of decoding.
                If `None` (default), zero state is used.
            max_decoding_length: A int scalar Tensor indicating the maximum
                allowed number of decoding steps. If `None` (default), either
                `hparams["max_decoding_length_train"]` or
                `hparams["max_decoding_length_infer"]` is used
                according to :attr:`mode`.
            impute_finished (bool): If `True`, then states for batch
                entries which are marked as finished get copied through and
                the corresponding outputs get zeroed out.  This causes some
                slowdown at each time step, but ensures that the final state
                and outputs have the correct values and that backprop ignores
                time steps that were marked as finished.
            helper (optional): An instance of
                :class:`~texar.torch.modules.Helper`
                that defines the decoding strategy. If given,
                ``decoding_strategy`` and helper configurations in
                :attr:`hparams` are ignored.

                :meth:`create_helper` can be used to create some of the common
                helpers for, e.g., teacher-forcing decoding, greedy decoding,
                sample decoding, etc.
            infer_mode (optional): If not `None`, overrides mode given by
                `self.training`.
            **kwargs: Other keyword arguments for constructing helpers
                defined by ``hparams["helper_train"]`` or
                ``hparams["helper_infer"]``.

        Returns:
            ``(outputs, final_state, sequence_lengths)``, where

            - `outputs`: an object containing the decoder output on all
              time steps.
            - `final_state`: the cell state of the final time step.
            - `sequence_lengths`: a :tensor:`LongTensor` of shape
              ``[batch_size]`` containing the length of each sample.
        """
        # TODO: Add faster code path for teacher-forcing training.

        # Helper
        if helper is None:
            helper = self._create_or_get_helper(infer_mode, **kwargs)

        if (isinstance(helper, decoder_helpers.TrainingHelper) and
                (inputs is None or sequence_length is None)):
            raise ValueError("'input' and 'sequence_length' must not be None "
                             "when using 'TrainingHelper'.")

        # Initial state
        self._cell.init_batch()

        # Maximum decoding length
        if max_decoding_length is None:
            if self.training:
                max_decoding_length = self._hparams.max_decoding_length_train
            else:
                max_decoding_length = self._hparams.max_decoding_length_infer
            if max_decoding_length is None:
                max_decoding_length = utils.MAX_SEQ_LENGTH

        return self.dynamic_decode(
            helper, inputs, sequence_length, initial_state,
            max_decoding_length, impute_finished)

    def _get_beam_search_cell(self):
        self._beam_search_cell = self._cell
        return self._cell

    @property
    def output_size(self):
        r"""Output size of one step.
        """
        raise NotImplementedError

    def initialize(self, helper: Helper, inputs: Optional[torch.Tensor],
                   sequence_length: Optional[torch.LongTensor],
                   initial_state: Optional[State]) \
            -> Tuple[torch.ByteTensor, torch.Tensor, Optional[State]]:
        initial_finished, initial_inputs = helper.initialize(
            self.embed_tokens, inputs, sequence_length)
        if initial_state is None:
            state = self._cell.init_batch()
        else:
            state = initial_state
        return (initial_finished, initial_inputs, state)

    def step(self, helper: Helper, time: int,
             inputs: torch.Tensor, state: Optional[State]) \
            -> Tuple[Output, State]:
        raise NotImplementedError

    def next_inputs(self, helper: Helper, time: int, outputs: Output) -> \
            Tuple[torch.Tensor, torch.ByteTensor]:
        raise NotImplementedError

    @property
    def cell(self):
        r"""The RNN cell.
        """
        return self._cell

    def zero_state(self, batch_size):
        r"""Zero state of the RNN cell.
        Equivalent to :attr:`decoder.cell.zero_state`.
        """
        return self._cell.zero_state(batch_size=batch_size)

    @property
    def state_size(self):
        r"""The state size of decoder cell.
        Equivalent to :attr:`decoder.cell.state_size`.
        """
        return self._cell.hidden_size

    @property
    def output_layer(self):
        r"""The output layer.
        """
        return self._output_layer



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for BERT encoders.
"""

import unittest

import torch

from texar.torch.modules.encoders.bert_encoder import BERTEncoder
from texar.torch.utils.test import pretrained_test


class BERTEncoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.BERTEncoder` class.
    """

    def setUp(self) -> None:
        self.batch_size = 2
        self.max_length = 3
        self.inputs = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_model_loading(self):
        r"""Tests model loading functionality."""
        for pretrained_model_name in BERTEncoder.available_checkpoints():
            encoder = BERTEncoder(pretrained_model_name=pretrained_model_name)
            _, _ = encoder(self.inputs)

    @pretrained_test
    def test_hparams(self):
        r"""Tests the priority of the encoder arch parameter.
        """
        # case 1: set "pretrained_mode_name" by constructor argument
        hparams = {
            "pretrained_model_name": "bert-large-uncased",
        }
        encoder = BERTEncoder(pretrained_model_name="bert-base-uncased",
                              hparams=hparams)
        self.assertEqual(encoder.hparams.encoder.num_blocks, 12)
        _, _ = encoder(self.inputs)

        # case 2: set "pretrained_mode_name" by hparams
        hparams = {
            "pretrained_model_name": "bert-large-uncased",
            "encoder": {
                "num_blocks": 6,
            }
        }
        encoder = BERTEncoder(hparams=hparams)
        self.assertEqual(encoder.hparams.encoder.num_blocks, 24)
        _, _ = encoder(self.inputs)

        # case 3: set to None in both hparams and constructor argument
        hparams = {
            "pretrained_model_name": None,
            "encoder": {
                "num_blocks": 6,
            },
        }
        encoder = BERTEncoder(hparams=hparams)
        self.assertEqual(encoder.hparams.encoder.num_blocks, 6)
        _, _ = encoder(self.inputs)

        # case 4: using default hparams
        encoder = BERTEncoder()
        self.assertEqual(encoder.hparams.encoder.num_blocks, 12)
        _, _ = encoder(self.inputs)

    @pretrained_test
    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        # case 1: bert base
        encoder = BERTEncoder()
        self.assertEqual(len(encoder.trainable_variables), 3 + 2 + 12 * 16 + 2)
        _, _ = encoder(self.inputs)

        # case 2: bert large
        hparams = {
            "pretrained_model_name": "bert-large-uncased"
        }
        encoder = BERTEncoder(hparams=hparams)
        self.assertEqual(len(encoder.trainable_variables), 3 + 2 + 24 * 16 + 2)
        _, _ = encoder(self.inputs)

        # case 3: self-designed bert
        hparams = {
            "encoder": {
                "num_blocks": 6,
            },
            "pretrained_model_name": None,
        }
        encoder = BERTEncoder(hparams=hparams)
        self.assertEqual(len(encoder.trainable_variables), 3 + 2 + 6 * 16 + 2)
        _, _ = encoder(self.inputs)

    def test_encode(self):
        r"""Tests encoding.
        """
        # case 1: bert base
        hparams = {
            "pretrained_model_name": None,
        }
        encoder = BERTEncoder(hparams=hparams)

        inputs = torch.randint(30521, (self.batch_size, self.max_length))
        outputs, pooled_output = encoder(inputs)

        outputs_dim = encoder.hparams.encoder.dim
        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, outputs_dim]))
        self.assertEqual(
            pooled_output.shape,
            torch.Size([self.batch_size, encoder.output_size]))

        # case 2: self-designed bert
        hparams = {
            'pretrained_model_name': None,
            'embed': {
                'dim': 96,
            },
            'segment_embed': {
                'dim': 96,
            },
            'position_embed': {
                'dim': 96,
            },

            'encoder': {
                'dim': 96,
                'multihead_attention': {
                    'num_units': 96,
                    'output_dim': 96,
                },
                'poswise_feedforward': {
                    'layers': [
                        {
                            'kwargs': {
                                'in_features': 96,
                                'out_features': 96 * 4,
                                'bias': True,
                            },
                            'type': 'Linear',
                        },
                        {"type": "BertGELU"},
                        {
                            'kwargs': {
                                'in_features': 96 * 4,
                                'out_features': 96,
                                'bias': True,
                            },
                            'type': 'Linear',
                        }
                    ]
                },
            },
            'hidden_size': 96,
        }
        encoder = BERTEncoder(hparams=hparams)

        outputs, pooled_output = encoder(inputs)

        outputs_dim = encoder.hparams.encoder.dim
        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, outputs_dim]))
        self.assertEqual(
            pooled_output.shape,
            torch.Size([self.batch_size, encoder.output_size]))

    def test_soft_ids(self):
        r"""Tests soft ids.
        """
        hparams = {
            "pretrained_model_name": None,
        }
        encoder = BERTEncoder(hparams=hparams)

        inputs = torch.rand(self.batch_size, self.max_length, 30522)
        outputs, pooled_output = encoder(inputs)

        outputs_dim = encoder.hparams.encoder.dim
        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, outputs_dim]))
        self.assertEqual(
            pooled_output.shape,
            torch.Size([self.batch_size, encoder.output_size]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from texar.torch.modules.encoders.transformer_encoder import TransformerEncoder
from texar.torch.modules.pretrained.t5_utils import \
    T5LayerNorm, MultiheadRPRAttention
from texar.torch.modules.networks.networks import FeedForwardNetwork
from texar.torch.modules.encoders.transformer_encoder import \
    default_transformer_poswise_net_hparams
from texar.torch.utils import sequence_mask, transformer_attentions as attn


class T5Encoder(TransformerEncoder):
    r"""Transformer based encoder that applies multi-head self attention with
     relative positional representations for encoding sequences for T5.

    This module basically stacks
    :class:`~texar.torch.modules.MultiheadRPRAttention`,
    :class:`~texar.torch.modules.FeedForwardNetwork` and residual connections.
    This module supports the standard T5 architecture proposed in
    `(Raffel et al.) "Exploring the Limits of Transfer Learning with a Unified
    Text-to-Text Transformer"`.

    Args:
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    .. document private functions
    """
    def __init__(self, hparams=None):
        super().__init__(hparams=hparams)

        self.final_layer_norm = T5LayerNorm(self._input_size,
                                            eps=self._hparams.eps)

    def initialize_blocks(self):
        r""" Helper function to initialize blocks.

        """
        for i in range(self._hparams.num_blocks):
            mh_attn = MultiheadRPRAttention(
                self._input_size,
                self._hparams.multihead_attention,
                stores_relative_position=bool(i == 0)
            )
            self.self_attns.append(mh_attn)

            self.self_attn_layer_norm.append(
                T5LayerNorm(self._input_size, eps=self._hparams.eps))
            if self._hparams.dim != mh_attn.hparams.output_dim:
                raise ValueError(
                    'The "dim" in the hparams of '
                    '"multihead_attention" should be equal to the '
                    '"dim" of T5Encoder')

            pw_net = FeedForwardNetwork(
                hparams=self._hparams['poswise_feedforward'])

            final_dim = pw_net.hparams.layers[-1]['kwargs']['out_features']
            if self._hparams.dim != final_dim:
                raise ValueError(
                    'The output dimenstion of '
                    '"poswise_feedforward" should be equal '
                    'to the "dim" of T5Encoder.')

            self.poswise_networks.append(pw_net)
            self.poswise_layer_norm.append(
                T5LayerNorm(self._input_size, eps=self._hparams.eps))

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "num_blocks": 6,
                "dim": 512,
                "embedding_dropout": 0.1,
                "residual_dropout": 0.1,
                "use_bert_config: False,
                "poswise_feedforward": default_transformer_poswise_net_hparams,
                'multihead_attention': {
                    'name': 'multihead_rpr_attention',
                    'num_units': 512,
                    'num_heads': 8,
                    'dropout_rate': 0.1,
                    'output_dim': 512,
                    'use_bias': False,
                    'is_decoder': False,
                    'relative_attention_num_buckets': 32
                },
                "initializer": None,
                "name": "t5_encoder"
            }

        Here:

        `"num_blocks"`: int
            Number of stacked blocks.

        `"dim"`: int
            Hidden dimension of the encoders.

        `"embedding_dropout"`: float
            Dropout rate of the input embedding.

        `"residual_dropout"`: float
            Dropout rate of the residual connections.

        "eps"`: float
            Epsilon values for layer norm layers.

        `"poswise_feedforward"`: dict
            Hyperparameters for a feed-forward network used in residual
            connections.
            Make sure the dimension of the output tensor is equal to ``"dim"``.
            See
            :func:`~texar.torch.modules.default_transformer_poswise_net_hparams`
            for details.

        `"multihead_attention"`: dict
            Hyperparameters for the multi-head attention strategy.
            Make sure the ``"output_dim"`` in this module is equal to ``"dim"``.
            See :class:`~texar.torch.modules.MultiheadRPRAttention` for
            details.

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"name"`: str
            Name of the module.
        """
        dim = 512
        return {
            'num_blocks': 6,
            'dim': dim,
            'embedding_dropout': 0.1,
            'residual_dropout': 0.1,
            'use_bert_config': False,
            'poswise_feedforward': default_transformer_poswise_net_hparams(dim),
            'multihead_attention': {
                'name': 'multihead_rpr_attention',
                'num_units': 512,
                'num_heads': 8,
                'dropout_rate': 0.1,
                'output_dim': 512,
                'use_bias': False,
                'is_decoder': False,
                'relative_attention_num_buckets': 32
            },
            'initializer': None,
            'eps': 1e-6,
            'name': 't5_encoder',
        }

    def forward(self,  # type: ignore
                inputs: torch.Tensor,
                sequence_length: torch.LongTensor) -> torch.Tensor:
        r"""Encodes the inputs.

        Args:
            inputs: A 3D Tensor of shape ``[batch_size, max_time, dim]``,
                containing the embedding of input sequences. Note that
                the embedding dimension `dim` must equal "dim" in
                :attr:`hparams`. The input embedding is typically an
                aggregation of word embedding and position embedding.
            sequence_length: A 1D :tensor:`LongTensor` of shape
                ``[batch_size]``. Input tokens beyond respective sequence
                lengths are masked out automatically.

        Returns:
            A Tensor of shape ``[batch_size, max_time, dim]`` containing the
            encoded vectors.
        """
        # Multiply input embedding with the sqrt of its dimension for
        # normalization

        inputs_padding = 1 - sequence_mask(
            sequence_length, inputs.size()[1]).float()
        ignore_padding = attn.attention_bias_ignore_padding(inputs_padding)
        encoder_self_attention_bias = ignore_padding

        x = self.embed_dropout(inputs)

        position_bias = None

        for i in range(self._hparams.num_blocks):
            _queries_input = self.self_attn_layer_norm[i](x)

            attention_output, position_bias = self.self_attns[i](
                queries=_queries_input,
                memory=_queries_input,
                memory_attention_bias=encoder_self_attention_bias,
                position_bias=position_bias
            )

            attention_output = self.residual_dropout(attention_output)

            x = x + attention_output

            poswise_network = self.poswise_networks[i]
            poswise_normalizer = self.poswise_layer_norm[i]

            y = poswise_normalizer(x)

            original_shape = y.size()

            y = y.view(-1, self._hparams.dim)

            layer_output = poswise_network(y)
            sub_output = self.residual_dropout(layer_output)
            sub_output = sub_output.view(original_shape)

            x = x + sub_output

        x = self.final_layer_norm(x)

        return x



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base data class that is inherited by all data classes.
A data defines data reading, parsing, batching, and other
preprocessing operations.
"""
import warnings
from abc import ABC
from typing import (
    Callable, Dict, Generic, Iterable, Iterator, List, Optional, Sequence,
    Tuple, TypeVar, Union)

import torch
from torch.utils.data import Dataset

from texar.torch.data.data.dataset_utils import Batch
from texar.torch.data.data.dataset_utils import _CacheStrategy, _LazyStrategy
from texar.torch.hyperparams import HParams

__all__ = [
    "DataSource",
    "SequenceDataSource",
    "IterDataSource",
    "ZipDataSource",
    "FilterDataSource",
    "RecordDataSource",
    "DatasetBase",
]

RawExample = TypeVar('RawExample')  # type of a raw example loaded from source
Example = TypeVar('Example')  # type of a data example


class DataSource(Generic[RawExample], ABC):
    r"""Base class for all data sources. A data source represents the *source*
    of the data, from which raw data examples are read and returned.

    Different to PyTorch :class:`~torch.utils.data.Dataset`, subclasses of this
    class are not required to implement :meth:`__getitem__` (default
    implementation raises `TypeError`), which is beneficial for certain sources
    that only supports iteration (reading from text files, reading Python
    iterators, etc.)
    """

    def __getitem__(self, index: int) -> RawExample:
        raise TypeError("This DataSource does not support random access")

    def __iter__(self) -> Iterator[RawExample]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise TypeError("This DataSource does not support random access")


class SequenceDataSource(DataSource[RawExample]):
    r"""Data source for reading from Python sequences.

    This data source supports indexing.

    Args:
        sequence: The Python sequence to read from. Note that a sequence should
            be iterable and supports `len`.
    """

    def __init__(self, sequence: Sequence[RawExample]):
        self._seq = sequence

    def __getitem__(self, index: int) -> RawExample:
        return self._seq[index]

    def __iter__(self) -> Iterator[RawExample]:
        return iter(self._seq)

    def __len__(self) -> int:
        return len(self._seq)


class IterDataSource(DataSource[RawExample]):
    r"""Data source for reading from Python iterables. Please note: if passed
    an *iterator* and caching strategy is set to 'none', then the data source
    can only be iterated over once.

    This data source does not support indexing.

    Args:
        iterable: The Python iterable to read from.
    """

    def __init__(self, iterable: Iterable[RawExample]):
        self._iter = iterable

    def __iter__(self) -> Iterator[RawExample]:
        return iter(self._iter)


class ZipDataSource(DataSource[Tuple[RawExample, ...]]):
    r"""Data source by combining multiple sources. The raw examples returned
    from this data source are tuples, with elements being raw examples from each
    of the constituting data sources.

    This data source supports indexing if all the constituting data sources
    support indexing.

    Args:
        sources: The list of data sources to combine.
    """

    def __init__(self, *sources: DataSource[RawExample]):
        self._sources = list(sources)

    def __getitem__(self, index: int) -> Tuple[RawExample, ...]:
        return tuple(source[index] for source in self._sources)

    def __iter__(self) -> Iterator[Tuple[RawExample, ...]]:
        return zip(*[iter(source) for source in self._sources])

    def __len__(self) -> int:
        return min(len(source) for source in self._sources)


class FilterDataSource(DataSource[RawExample]):
    r"""Data source for filtering raw examples with a user-specified filter
    function. Only examples for which the filter functions returns `True` are
    returned.

    This data source supports indexing if the wrapped data source supports
    indexing.

    Args:
        source: The data source to filter.
        filter_fn: A callable taking a raw example as argument and returning a
            boolean value, indicating whether the raw example should be
            **kept**.
    """

    def __init__(self, source: DataSource[RawExample],
                 filter_fn: Callable[[RawExample], bool]):
        self._source = source
        self._filter_fn = filter_fn

    def __iter__(self) -> Iterator[RawExample]:
        for sentence in self._source:
            if self._filter_fn(sentence):
                yield sentence


class RecordDataSource(DataSource[Dict[str, RawExample]]):
    r"""Data source by structuring multiple sources. The raw examples returned
    from this data source are dictionaries, with values being raw examples from
    each of the constituting data sources.

    This data source supports indexing if all the constituting data sources
    support indexing.

    Args:
        sources: A dictionary mapping names to data sources, containing the
            data sources to combine.
    """

    def __init__(self, sources: Dict[str, DataSource[RawExample]]):
        self._sources = sources

    def __getitem__(self, index: int) -> Dict[str, RawExample]:
        return {key: source[index] for key, source in self._sources.items()}

    def __iter__(self) -> Iterator[Dict[str, RawExample]]:
        keys = list(self._sources.keys())
        iterator = zip(*[iter(source) for source in self._sources.values()])
        for values in iterator:
            yield dict(zip(keys, values))

    def __len__(self) -> int:
        return min(len(source) for source in self._sources.values())


class _TruncatedDataSource(DataSource[RawExample]):
    def __init__(self, data_source: DataSource[RawExample], max_size: int):
        self._source = data_source
        self._max_size = max_size

    def __getitem__(self, item) -> RawExample:
        if item >= self._max_size:
            raise IndexError(
                f"Data index ({item}) out of range [0, {self._max_size})")
        return self._source[item]

    def __iter__(self) -> Iterator[RawExample]:
        count = 0
        iterator = iter(self._source)
        while count < self._max_size:
            yield next(iterator)
            count += 1

    def __len__(self) -> int:
        try:
            length = min(len(self._source), self._max_size)
        except TypeError:
            length = self._max_size
        return length


class _TransformedDataSource(DataSource[Example], Generic[RawExample, Example]):
    r"""Data source by performing transformations on another data source.
    """

    def __init__(self, data_source: DataSource[RawExample],
                 process_fn: Callable[[RawExample], Example]):
        self._source = data_source
        self._process = process_fn

    def __getitem__(self, item):
        return self._process(self._source[item])

    def __iter__(self):
        return map(self._process, iter(self._source))

    def __len__(self):
        return len(self._source)

    def __getattr__(self, item):
        return getattr(self._source, item)


class _CachedDataSource(DataSource[RawExample]):
    r"""Wrapper for random access support over a data source that does not
    implement `__getitem__`. This class is only used internally in
    :class:`~texar.torch.data.data.DatasetBase`, while conforming to user
    `cache_strategy` and `shuffle_buffer_size` settings.
    """

    _cache: Union[Dict[int, RawExample], List[RawExample]]

    def __init__(self, data_source: DataSource[RawExample],
                 erase_after_access: bool = True):
        r"""

        Args:
            data_source: The data source to wrap around.
            erase_after_access: If `True`, cached examples are erased after
                being accessed through `__getitem__`. Useful when
                :class:`~texar.torch.data.data.DatasetBase` hyperparameter
                `cache_strategy` is set to `none` or `processed`.
        """
        self._source = data_source
        self._iter = iter(data_source)
        self._max_index = -1
        self._erase_after_access = erase_after_access
        if erase_after_access:
            self._cache: Dict[int, RawExample] = {}
        else:
            self._cache: List[RawExample] = []

    def __getitem__(self, index: int) -> RawExample:
        # If specified `index` is not yet prefetched (or has already been
        # accessed), this method may throw `IndexError` or `KeyError`.
        example = self._cache[index]
        if self._erase_after_access:
            del self._cache[index]
        return example

    def __iter__(self) -> Iterator[RawExample]:
        return iter(self._source)

    def prefetch(self, index: int):
        while self._max_index < index:
            example = next(self._iter)
            self._max_index += 1
            if self._erase_after_access:
                self._cache[self._max_index] = example
            else:
                self._cache.append(example)  # type: ignore

    @property
    def max_index(self) -> int:
        return self._max_index

    def reset(self) -> None:
        if self._erase_after_access:
            self._iter = iter(self._source)
            self._max_index = -1


class DatasetBase(Dataset, Generic[RawExample, Example], ABC):
    r"""Base class inherited by all data classes.

    Args:
        source: An instance of type :class:`~texar.torch.data.DataSource`,
        hparams: A `dict` or instance of :class:`~texar.torch.HParams`
            containing hyperparameters. See :meth:`default_hparams` for the
            defaults.
        device: The device of the produced batches. For GPU training, set to
            current CUDA device.

            .. note::
                When :attr:`device` is set to a CUDA device, tensors in the
                batch will be automatically moved to the specified device. This
                may result in performance issues if your data examples contain
                complex structures (e.g., nested lists with many elements). In
                this case, it is recommended to set :attr:`device` to `None` and
                manually move your data.

                For more details, see :meth:`collate`.

    Users can also directly inherit from this class to implement customized data
    processing routines. Two methods should be implemented in the subclass:

    - :meth:`process`: Process a single data example read from the data source
      (*raw example*). Default implementation returns the raw example as is.
    - :meth:`collate`: Combine a list of processed examples into a single batch,
      and return an object of type :class:`~texar.torch.data.Batch`.

    Example:

        Here, we define a custom data class named ``MyDataset``, which is
        equivalent to the most basic usage of
        :class:`~texar.torch.data.MonoTextData`.

        .. code-block:: python

            class MyDataset(tx.data.DatasetBase):
                def __init__(self, data_path, vocab, hparams=None, device=None):
                    source = tx.data.TextLineDataSource(data_path)
                    self.vocab = vocab
                    super().__init__(source, hparams, device)

                def process(self, raw_example):
                    # `raw_example` is a data example read from `self.source`,
                    # in this case, a line of tokenized text, represented as a
                    # list of `str`.
                    return {
                        "text": raw_example,
                        "ids": self.vocab.map_tokens_to_ids_py(raw_example),
                    }

                def collate(self, examples):
                    # `examples` is a list of objects returned from the
                    # `process` method. These data examples should be collated
                    # into a batch.

                    # `text` is a list of list of `str`, storing the tokenized
                    # sentences for each example in the batch.
                    text = [ex["text"] for ex in examples]
                    # `ids` is the NumPy tensor built from the token IDs of each
                    # sentence, and `lengths` the lengths of each sentence.
                    # The `tx.data.padded_batch` function pads IDs to the same
                    # length and then stack them together. This function is
                    # commonly used in `collate` methods.
                    ids, lengths = tx.data.padded_batch(
                        [ex["ids"] for ex in examples])
                    return tx.data.Batch(
                        len(examples),
                        text=text,
                        text_ids=torch.from_numpy(ids),
                        lengths=torch.tensor(lengths))

            vocab = tx.data.Vocab("vocab.txt")
            hparams = {'batch_size': 1}
            data = MyDataset("data.txt", vocab, hparams)
            iterator = DataIterator(data)
            for batch in iterator:
                # batch contains the following
                # batch_ == {
                #    'text': [['<BOS>', 'example', 'sequence', '<EOS>']],
                #    'text_ids': [[1, 5, 10, 2]],
                #    'length': [4]
                # }
    """

    # pylint: disable=line-too-long

    # The `DatasetBase` is used in combination with Texar `DataIterator`, which internally uses the PyTorch `DataLoader`
    # for multi-processing support.
    #
    # We divide the entire data pipeline into three stages, namely *load*, *process*, and *batch*:
    # - **Load** refers to loading data from the data source (e.g., a file, a Python list or iterator). In Texar,
    #   loading is handled by `DataSource` classes.
    # - **Process** refers to preprocessing routines for each data example (e.g., vocabulary mapping, tokenization). In
    #   Texar, this is the `process` function of each `DatasetBase` class.
    # - **Batch** refers to combining multiple examples to form a batch, which typically includes padding and moving
    #   data across devices. In Texar, this is the `collate` function of each `DatasetBase` class.
    #
    # PyTorch DataLoader only performs batching, and since multi-processing is used, the entire dataset is expected to
    # be in memory before iteration, i.e. loading and processing cannot be lazy. The `DatasetBase` class is carefully
    # crafted to provide laziness and caching options at all possible stages.
    #
    # To support laziness, we pass data examples (either raw or processed, depending on whether processing is lazy) to
    # the worker processes. To prevent modifying the underlying `DataLoader` implementation, we hack the PyTorch
    # `Sampler` classes (responsible for sampling the next data example from the dataset, and returning its index) to
    # also return data examples. To support caching, the worker may also need to return the processed examples through
    # pipes.
    #
    # The following table describes the intended behavior of each combination of lazy/caching modes, and the exact
    # behaviors of the sampler and workers. `<X>` means the mode combination does not make sense (e.g. with `Lazy.None`,
    # processed data examples are effectively cached, so `Cache.None` makes no sense). Parts in `*[blah]*` hold true
    # only for the first epoch.
    #
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    # |               | Cache.None                    | Cache.Loaded                  | Cache.Processed               |
    # |               | no caching                    | only cache loaded examples    | only cache processed examples |
    # +===============+===============================+===============================+===============================+
    # | Lazy.None     | <X>                           | <X>                           | Sampler returns indices.      |
    # | eager load,   |                               |                               | Worker only does batching.    |
    # | eager process |                               |                               | Worker returns batch.         |
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    # | Lazy.Process  | <X>                           | Sampler returns indices.      | Sampler returns indices.      |
    # | eager load,   |                               | Worker does batching and      | Worker does batching          |
    # | lazy process  |                               |   processing.                 |   *[and processing]*.         |
    # |               |                               | Worker returns batch.         | Worker returns batch          |
    # |               |                               |                               |   *[and processed examples]*. |
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    # | Lazy.All      | Sampler returns indices and   | Sampler returns indices       | Sampler returns indices       |
    # | lazy load,    |   data examples.              |   *[and data examples]*.      |   *[and data examples]*.      |
    # | lazy process  | Worker does batching and      | Worker does batching and      | Worker does batching          |
    # |               |   processing.                 |   processing.                 |   *[and processing]*.         |
    # |               | Worker returns batch.         | Worker returns batch.         | Worker returns batch          |
    # |               |                               |                               |   *[and processed examples]*. |
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    #
    # Note that in the above table we assume `parallelize_processing` to be True. In rare cases this may not be desired,
    # for instance, when `process` depends on some shared variable that must be modified during iteration, e.g. a
    # vocabulary constructed on-the-fly. When `parallelize_processing` is False, behaviors are as the following (much
    # simpler) table. Although, note that compared to the above cases, this often results in worse performance.
    #
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    # |               | Cache.None                    | Cache.Loaded                  | Cache.Processed               |
    # |               | no caching                    | only cache loaded examples    | only cache processed examples |
    # +===============+===============================+===============================+===============================+
    # | Lazy.None     | <X>                           | <X>                           | Sampler returns indices.      |
    # | eager load,   |                               |                               | Worker only does batching.    |
    # | eager process |                               |                               | Worker returns batch.         |
    # +---------------+-------------------------------+-------------------------------+-------------------------------+
    # | Lazy.Process  | <X>                           | Sampler returns indices and processed examples.               |
    # | eager load,   |                               | Worker only does batching.                                    |
    # | lazy process  |                               | Worker returns batch.                                         |
    # +---------------+-------------------------------+---------------------------------------------------------------+
    # | Lazy.All      | Sampler returns indices and processed examples.                                               |
    # | lazy load,    | Worker only does batching.                                                                    |
    # | lazy process  | Worker returns batch.                                                                         |
    # +---------------+-----------------------------------------------------------------------------------------------+

    # pylint: enable=line-too-long

    _source: DataSource[RawExample]
    _dataset_size: Optional[int]

    def __init__(self, source: DataSource[RawExample], hparams=None,
                 device: Optional[torch.device] = None):
        self._source = source
        self._hparams = HParams(hparams, self.default_hparams())
        self.device = device

        if self._hparams.num_epochs != 1:
            warnings.warn(f"'num_epochs' is set to {self._hparams.num_epochs}, "
                          f"but will be treated as 1.")

        # Check and convert strategy hyperparameters.
        self._lazy_strategy = _LazyStrategy(self._hparams.lazy_strategy)
        self._cache_strategy = _CacheStrategy(self._hparams.cache_strategy)
        if self._lazy_strategy is _LazyStrategy.NONE:
            if self._cache_strategy is not _CacheStrategy.PROCESSED:
                warnings.warn(
                    f"Using '{self._cache_strategy}' cache strategy with "
                    f"'none' lazy strategy. This will be equivalent to "
                    f"'processed' cache strategy.")
            self._cache_strategy = _CacheStrategy.PROCESSED
        elif self._lazy_strategy is _LazyStrategy.PROCESS:
            if self._cache_strategy is _CacheStrategy.NONE:
                warnings.warn(
                    f"Using 'none' cache strategy with 'process' lazy "
                    f"strategy. This will be equivalent to 'loaded' cache "
                    f"strategy.")
                self._cache_strategy = _CacheStrategy.LOADED
        self._uses_multi_processing = self._hparams.num_parallel_calls > 0
        self._parallelize_processing = self._hparams.parallelize_processing

        self._processed_cache: List[Example] = []
        self._fully_cached = False

        # If specified maximum dataset size, wrap the data source. This is done
        # before caching to avoid caching excess elements.
        if self._hparams.max_dataset_size != -1:
            self._source = _TruncatedDataSource[RawExample](
                self._source, self._hparams.max_dataset_size)

        # If processing should not be parallelized, combine processing with
        # loading by wrapping the data source. In this case, **processed** data
        # will be cached.
        if (not self._parallelize_processing and
                self._lazy_strategy is _LazyStrategy.ALL and
                self._cache_strategy is not _CacheStrategy.LOADED):
            self._transformed_source = _TransformedDataSource[
                RawExample, Example](self._source, self.process)
            self._source = self._transformed_source  # type: ignore

        # Check whether data source supports random access, and obtain dataset
        # size if it does.
        self._supports_random_access = True
        if self._lazy_strategy is not _LazyStrategy.NONE:
            try:
                self._dataset_size = len(self._source)
                _ = self._source[0]
            except TypeError:
                self._supports_random_access = False
                erase_after_access = (
                        self._cache_strategy is not _CacheStrategy.LOADED)
                self._cached_source = _CachedDataSource[RawExample](
                    self._source, erase_after_access)
                self._source = self._cached_source
                self._dataset_size = None

        # If processing should not be parallelized, combine processing with
        # loading by wrapping the data source. In this case, **loaded** data
        # will be cached.
        if (not self._parallelize_processing and
                self._cache_strategy is _CacheStrategy.LOADED):
            self._transformed_source = _TransformedDataSource[
                RawExample, Example](self._source, self.process)
            self._source = self._transformed_source  # type: ignore

        # Simplify some logic-heavy checks.
        self.__should_return_processed_examples = (
                self._lazy_strategy is not _LazyStrategy.NONE and
                self._cache_strategy is _CacheStrategy.PROCESSED and
                self._parallelize_processing)
        self.__should_call_prefetch_source = (
                self._lazy_strategy is _LazyStrategy.ALL and
                self._cache_strategy is _CacheStrategy.NONE)
        self.__should_call_prefetch_processed = (
                not self._parallelize_processing and
                self._lazy_strategy is _LazyStrategy.PROCESS and
                self._cache_strategy is _CacheStrategy.PROCESSED)
        self.__should_delete_source_in_add_cache = (
                not self._supports_random_access and
                self._parallelize_processing and
                self._uses_multi_processing and
                self._lazy_strategy is _LazyStrategy.PROCESS and
                self._cache_strategy is _CacheStrategy.PROCESSED)

        # Perform eager loading/processing if required.
        if self._lazy_strategy is _LazyStrategy.NONE:
            # Process entire dataset and cache.
            self._processed_cache = [self.process(raw_example)
                                     for raw_example in self._source]
            self._dataset_size = len(self._processed_cache)
            self._fully_cached = True
        else:
            if self._lazy_strategy is _LazyStrategy.PROCESS:
                # Load entire dataset. Note that if data source supports random
                # access, we assume it is already loaded into memory.
                if not self._supports_random_access:
                    self._prefetch_all_source()

            if self._cache_strategy is _CacheStrategy.PROCESSED:
                # Data can be processed in arbitrary order, so they need to be
                # reordered before storing in the cache list.
                self._reorder_cache: Dict[int, Example] = {}

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of default hyperparameters.

        .. code-block:: python

            {
                "num_epochs": 1,
                "batch_size": 64,
                "allow_smaller_final_batch": True,
                "shuffle": True,
                "shuffle_buffer_size": None,
                "shard_and_shuffle": False,
                "num_parallel_calls": 1,
                "prefetch_buffer_size": 0,
                "max_dataset_size": -1,
                "seed": None,
                "lazy_strategy": 'none',
                "cache_strategy": 'processed',
                "parallelize_processing": True,
                "name": "data"
            }

        Here:

        `"num_epochs"`: int
            Number of times the dataset should be repeated.

            .. note::
                This option only exists for compatibility, and will be
                ignored. A warning will be generated is any value other than
                1 is used.

        `"batch_size"`: int
            Batch size, i.e., the number of consecutive elements of the
            dataset to combine in a single batch.

        `"allow_smaller_final_batch"`: bool
           Whether to allow the final batch to be smaller if there are
           insufficient elements left. If `False`, the final batch is
           discarded if it is smaller than batch size. Note that,
           if `True`, `output_shapes` of the resulting dataset
           will have a a **static** batch_size dimension equal to
           "batch_size".

        `"shuffle"`: bool
            Whether to randomly shuffle the elements of the dataset.

        `"shuffle_buffer_size"`: int
            The buffer size for data shuffling. The larger, the better
            the resulting data is mixed.

            If `None` (default), buffer size is set to the size of the
            whole dataset (i.e., make the shuffling the maximally
            effective).

        `"shard_and_shuffle"`: bool
            Whether to first shard the dataset and then shuffle each
            block respectively. Useful when the whole data is too large to
            be loaded efficiently into the memory.

            If `True`, :attr:`shuffle_buffer_size` must be specified to
            determine the size of each shard.

            .. warning::
                Sharding is not yet supported. This option will be ignored.

        `"num_parallel_calls"`: int
            Number of elements from the datasets to process in parallel.
            When ``"num_parallel_calls"`` equals 0, no worker processes will
            be created; when the value is greater than 0, the number of worker
            processes will be equal to ``"num_parallel_calls"``.

        `"prefetch_buffer_size"`: int
            The maximum number of elements that will be buffered when
            prefetching.

            .. note::
                This option exists only for compatibility. Currently data
                is only prefetched when ``"num_parallel_calls"`` is greater
                than 1, and the number of examples to prefetch is controlled
                internally by PyTorch :torch_docs:`DataLoader
                <data.html#torch.utils.data.DataLoader>`.

        `"max_dataset_size"`: int
            Maximum number of instances to include in
            the dataset. If set to `-1` or greater than the size of
            dataset, all instances will be included. This constraint is
            imposed after data shuffling and filtering.

        `"seed"`: int, optional
            The random seed for shuffle.

            Note that if a seed is set, the shuffle order will be exact
            the same every time when going through the (repeated) dataset.

            .. warning::
                Manual seeding is not yet supported. This option will be
                ignored.

        `"lazy_strategy"`: str
            Lazy strategy for data examples. Lazy loading/processing defers
            data loading/processing until when it's being accessed.
            Non-lazy (eager) loading/processing would load/process all data
            upon construction of dataset. Available options are:

            - `none`: Perform eager loading and processing.
            - `process`: Perform eager loading and lazy processing.
            - `all`: Perform lazy loading and processing.

            Defaults to `all`. Note that currently, all eager operations
            are performed on a single process only.

        `"cache_strategy"`: str
            Caching strategy for data examples. Available options are:

            - `none`: No data is cached. Data is always loaded from
              source (e.g. file) and processed upon access.
            - `loaded`: Only cache raw data loaded from source,
              processing routines are performed upon access.
            - `processed`: Processed data is cached. **Note:** raw data
              will not be cached in this case, because raw data is
              only used to construct the processed data.

            Default value is `loaded`. This option depends on the value of
            `lazy_strategy`, specifically:

            - When `lazy_strategy` is `none`, all choices of
              `cache_strategy` are equivalent to `processed`.
            - When `lazy_strategy` is `process`, `none` is equivalent
              to `loaded`.

        `"parallelize_processing"`: bool
            Whether to perform parallelized processing of data. Since
            multi-processing parallelism is utilized, this flag should be
            `False` if your process routine involves modifying a shared
            object across examples.

            Note that this only affects cases where `lazy_strategy` is not
            `none`. If `lazy_strategy` is `none`, processing will be
            performed on a single process regardless of this value.

        `"name"`: str
            Name of the data.
        """
        # TODO: Sharding not yet supported.
        # TODO: `seed` is not yet applied.
        # TODO: `prefetch_buffer_size` will not be supported, but could remain
        #   for compatibility.
        return {
            "name": "data",
            "num_epochs": 1,
            "batch_size": 64,
            "allow_smaller_final_batch": True,
            "shuffle": True,
            "shuffle_buffer_size": None,
            "shard_and_shuffle": False,
            "num_parallel_calls": 0,
            "prefetch_buffer_size": 0,
            "max_dataset_size": -1,
            "seed": None,
            "lazy_strategy": 'none',
            "cache_strategy": 'processed',
            "parallelize_processing": True,
        }

    def to(self, device: Optional[torch.device]):
        r"""Move the dataset to the specific device. Note that we don't actually
        move data or do anything here -- data will be moved to the appropriate
        device after :class:`~texar.torch.data.DataIterator` fetches the batch.
        """
        if device is not None:
            self.device = device
        return self

    def _prefetch_processed(self, index: int):
        r"""Performs processing on the main process. This is called in
        :meth:`texar.torch.data.data.DatasetBase._prefetch_source` if
        `parallelize_processing` is `False`."""
        if len(self._processed_cache) <= index:
            self._processed_cache.extend(
                self.process(self._source[x])
                for x in range(len(self._processed_cache), index + 1))
            if len(self._processed_cache) == self._dataset_size:
                self._fully_cached = True

    def _prefetch_all_source(self) -> int:
        r"""Prefetches all examples from data source. This is only called if
        `__len__` is called before dataset size can be determined, or when using
        eager loading.
        """

        try:
            max_index = 10 ** 8
            self._cached_source.prefetch(max_index)
            warnings.warn(
                f"The data source contains more than {max_index:.2e} "
                f"examples. Please check whether it is infinite.")
            while True:
                max_index *= 2
                self._cached_source.prefetch(max_index)
        except StopIteration:
            self._dataset_size = self._cached_source.max_index + 1
            return self._dataset_size

    def _prefetch_source(self, index: int) -> Optional[int]:
        r"""Prefetches data so `__getitem__` will be available. This method
        should only be called in the main process, because data sources are not
        guaranteed to be thread-safe.

        Args:
            index: Prefetch data up to this index.

        Returns:
            If `index` is greater than dataset size, returns the inferred
            dataset size. Otherwise, returns `None`.
        """
        if not self._supports_random_access:
            try:
                self._cached_source.prefetch(index)
            except StopIteration:
                self._dataset_size = self._cached_source.max_index + 1
                # self._cached_source.reset()
                if self._should_call_prefetch_processed:
                    self._prefetch_processed(self._dataset_size - 1)
                return self._dataset_size
            if self._should_call_prefetch_processed:
                self._prefetch_processed(index)
        else:
            # Dataset size must be known.
            if index >= self._dataset_size:  # type: ignore
                return self._dataset_size
        return None

    def __len__(self) -> int:
        if self._dataset_size is None:
            raise TypeError(
                "__len__ not supported for datasets with undetermined size")
        return self._dataset_size

    def process(self, raw_example: RawExample) -> Example:
        r"""The process routine. A default implementation of no-op is provided,
        but subclasses are free to override this behavior.

        The process routine would take raw examples loaded from the data source
        as input, and return processed examples. If `parallelize_processing`
        is `True`, this method **must not** access shared variables that are
        modified during iterator (e.g., constructing vocabularies on-the-fly).

        Args:
            raw_example: The raw example loaded from data.

        Returns:
            The processed example.
        """
        return raw_example  # type: ignore

    def __getitem__(self, index: Union[int, Tuple[int, RawExample]]) -> Example:
        if isinstance(index, int):
            if self._fully_cached:
                return self._processed_cache[index]
            elif not self._parallelize_processing:
                return self._transformed_source[index]
            else:
                return self.process(self._source[index])
        else:
            # `index` is a tuple of (index, example).
            if not self._parallelize_processing:
                return index[1]  # type: ignore
            else:
                return self.process(index[1])

    def _add_cached_examples(self, indices: List[int], examples: List[Example]):
        r"""Called by :class:`texar.torch.data.data._CacheDataLoaderIter` to
        cache examples processed in worker processes.

        Args:
            indices: Indices for each example.
            examples: The examples processed in worker processes.
        """
        if self._should_delete_source_in_add_cache:
            # In this case, `_CachedDataSource.__getitem__` will be
            # called on worker processes, so the cache cannot be
            # deleted. Thus, we move deletion to
            # `_add_cached_examples`.
            for index in indices:
                del self._cached_source._cache[index]  # pylint: disable=protected-access
        for index, example in zip(indices, examples):
            if index == len(self._processed_cache):
                self._processed_cache.append(example)
            else:
                self._reorder_cache[index] = example

        while len(self._processed_cache) in self._reorder_cache:
            index = len(self._processed_cache)
            self._processed_cache.append(self._reorder_cache[index])
            del self._reorder_cache[index]
        if len(self._processed_cache) == self._dataset_size:
            self._fully_cached = True

    def _start_iteration(self) -> None:
        r"""Called by :class:`texar.torch.data.data.SamplerBase` before a new
        round of iteration starts. Note that this method will only be called if
        an unknown-sized iterator is used.
        """
        if not self._supports_random_access:
            self._cached_source.reset()

    @property
    def num_epochs(self):
        r"""Number of epochs.
        """
        return self._hparams.num_epochs

    @property
    def batch_size(self):
        r"""The batch size.
        """
        return self._hparams.batch_size

    @property
    def hparams(self):
        r"""A :class:`~texar.torch.HParams` instance of the
        data hyperparameters.
        """
        return self._hparams

    @property
    def name(self):
        r"""Name of the module.
        """
        return self._hparams.name

    @property
    def dataset(self):  # TODO: maybe change this to `data_source`
        r"""The data source.
        """
        return self._source

    def collate(self, examples: List[Example]) -> Batch:
        r"""The collate routine. Subclasses must implement this method.

        The collate routine is called to collate (combine) examples into
        batches. This function takes a list of processed examples, and returns
        an instance of :class:`~texar.torch.data.Batch`.

        .. note::
            Implementation should make sure that the returned callable is safe
            and efficient under multi-processing scenarios. Basically, do not
            rely on variables that could be modified during iteration, and avoid
            accessing unnecessary variables, as each access would result in a
            cross-process memory copy.

        .. warning::
            The recommended pattern is not to move tensor storage within this
            method, but you are free to do so.

            However, if multiple workers are used
            (:attr:`num_parallel_calls` > 0), moving tensors to CUDA devices
            within this method would result in CUDA errors being thrown.

        Args:
            examples: A list of processed examples in a batch.

        Returns:
            The collated batch.
        """
        raise NotImplementedError

    def _collate_and_maybe_return(self, examples: List[Example]) -> \
            Union[Batch, Tuple[List[Example], Batch]]:
        r"""Called by :class:`~texar.torch.data.DataIterator` to obtain the
        collated batch (and processed examples under certain circumstances).

        Args:
            examples: A list of processed examples in a batch.

        Returns:
            The collated batch.
        """
        batch = self.collate(examples)
        if self._should_return_processed_examples:
            return examples, batch
        return batch

    @property
    def _should_return_processed_examples(self):
        r"""Returns `True` if the worker threads should perform processing and
        return the processed examples.
        """
        return (not self._fully_cached and
                self.__should_return_processed_examples)

    @property
    def _should_yield_raw_example(self):
        r"""Returns `True` if the sampler should yield raw examples.
        """
        return (self._lazy_strategy is _LazyStrategy.ALL and
                (self._cache_strategy is _CacheStrategy.NONE or
                 not self._fully_cached))

    @property
    def _should_call_prefetch_source(self):
        r"""Returns `True` if the sampler should call `_prefetch_source`.
        """
        return (self._dataset_size is None or
                self.__should_call_prefetch_source)

    @property
    def _should_call_prefetch_processed(self):
        r"""Returns `True` if `_prefetch_source` should call
        `_prefetch_processed`.
        """
        return self.__should_call_prefetch_processed

    @property
    def _should_delete_source_in_add_cache(self):
        r"""Returns `True` if `_add_cached_examples` should delete cached raw
        examples.
        """
        return self.__should_delete_source_in_add_cache



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utils of RoBERTa Modules.
"""

import os
from abc import ABC
from typing import Any, Dict

import torch

from texar.torch.modules.pretrained.pretrained_base import PretrainedMixin

__all__ = [
    "PretrainedRoBERTaMixin",
]

_ROBERTA_PATH = "https://dl.fbaipublicfiles.com/fairseq/models/"


class PretrainedRoBERTaMixin(PretrainedMixin, ABC):
    r"""A mixin class to support loading pre-trained checkpoints for modules
    that implement the RoBERTa model.

    The RoBERTa model was proposed in (`Liu et al`. 2019)
    `RoBERTa: A Robustly Optimized BERT Pretraining Approach`_.
    As a variant of the standard BERT model, RoBERTa trains for more
    iterations on more data with a larger batch size as well as other tweaks
    in pre-training. Differing from the standard BERT, the RoBERTa model
    does not use segmentation embedding. Available model names include:

      * ``roberta-base``: RoBERTa using the BERT-base architecture,
        125M parameters.
      * ``roberta-large``: RoBERTa using the BERT-large architecture,
        355M parameters.

    We provide the following RoBERTa classes:

      * :class:`~texar.torch.modules.RoBERTaEncoder` for text encoding.
      * :class:`~texar.torch.modules.RoBERTaClassifier` for text
        classification and sequence tagging.

    .. _`RoBERTa: A Robustly Optimized BERT Pretraining Approach`:
        https://arxiv.org/abs/1907.11692
    """

    _MODEL_NAME = "RoBERTa"
    _MODEL2URL = {
        'roberta-base':
            _ROBERTA_PATH + "roberta.base.tar.gz",
        'roberta-large':
            _ROBERTA_PATH + "roberta.large.tar.gz",
    }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str) -> Dict[str, Any]:
        info = list(os.walk(cache_dir))
        root, _, files = info[0]
        config_path = None

        for file in files:
            if file.endswith('model.pt'):
                config_path = os.path.join(root, file)
                args = torch.load(config_path, map_location="cpu")['args']
                hidden_dim = args.encoder_embed_dim
                vocab_size = 50265
                position_size = args.max_positions + 2
                embedding_dropout = args.dropout
                num_blocks = args.encoder_layers
                num_heads = args.encoder_attention_heads
                dropout_rate = args.attention_dropout
                residual_dropout = args.dropout
                intermediate_size = args.encoder_ffn_embed_dim
                hidden_act = args.activation_fn

        if config_path is None:
            raise ValueError(f"Cannot find the config file in {cache_dir}")

        configs = {
            'hidden_size': hidden_dim,
            'embed': {
                'name': 'word_embeddings',
                'dim': hidden_dim
            },
            'vocab_size': vocab_size,
            'position_embed': {
                'name': 'position_embeddings',
                'dim': hidden_dim
            },
            'position_size': position_size,
            'encoder': {
                'name': 'encoder',
                'embedding_dropout': embedding_dropout,
                'num_blocks': num_blocks,
                'multihead_attention': {
                    'use_bias': True,
                    'num_units': hidden_dim,
                    'num_heads': num_heads,
                    'output_dim': hidden_dim,
                    'dropout_rate': dropout_rate,
                    'name': 'self'
                },
                'residual_dropout': residual_dropout,
                'dim': hidden_dim,
                'eps': 1e-12,
                'use_bert_config': True,
                'poswise_feedforward': {
                    "layers": [{
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': hidden_dim,
                            'out_features': intermediate_size,
                            'bias': True,
                        }
                    }, {
                        'type': 'Bert' + hidden_act.upper()
                    }, {
                        'type': 'Linear',
                        'kwargs': {
                            'in_features': intermediate_size,
                            'out_features': hidden_dim,
                            'bias': True,
                        }
                    }],
                },
            }
        }

        return configs

    def _init_from_checkpoint(self, pretrained_model_name: str,
                              cache_dir: str, **kwargs):
        global_tensor_map = {
            'decoder.sentence_encoder.embed_tokens.weight':
                'word_embedder._embedding',
            'decoder.sentence_encoder.embed_positions.weight':
                'position_embedder._embedding',
            'decoder.sentence_encoder.emb_layer_norm.weight':
                'encoder.input_normalizer.weight',
            'decoder.sentence_encoder.emb_layer_norm.bias':
                'encoder.input_normalizer.bias',
        }

        attention_tensor_map = {
            'final_layer_norm.weight':
                'encoder.output_layer_norm.{}.weight',
            'final_layer_norm.bias':
                'encoder.output_layer_norm.{}.bias',
            'fc1.weight':
                'encoder.poswise_networks.{}._layers.0.weight',
            'fc1.bias':
                'encoder.poswise_networks.{}._layers.0.bias',
            'fc2.weight':
                'encoder.poswise_networks.{}._layers.2.weight',
            'fc2.bias':
                'encoder.poswise_networks.{}._layers.2.bias',
            'self_attn_layer_norm.weight':
                'encoder.poswise_layer_norm.{}.weight',
            'self_attn_layer_norm.bias':
                'encoder.poswise_layer_norm.{}.bias',
            'self_attn.out_proj.weight':
                'encoder.self_attns.{}.O_dense.weight',
            'self_attn.out_proj.bias':
                'encoder.self_attns.{}.O_dense.bias',
            'self_attn.in_proj_weight': [
                'encoder.self_attns.{}.Q_dense.weight',
                'encoder.self_attns.{}.K_dense.weight',
                'encoder.self_attns.{}.V_dense.weight',
            ],
            'self_attn.in_proj_bias': [
                'encoder.self_attns.{}.Q_dense.bias',
                'encoder.self_attns.{}.K_dense.bias',
                'encoder.self_attns.{}.V_dense.bias'
            ],
        }

        checkpoint_path = os.path.abspath(os.path.join(cache_dir, 'model.pt'))
        device = next(self.parameters()).device
        params = torch.load(checkpoint_path, map_location=device)['model']

        for name, tensor in params.items():
            if name in global_tensor_map:
                v_name = global_tensor_map[name]
                pointer = self._name_to_variable(v_name)
                assert pointer.shape == tensor.shape
                pointer.data = tensor.data.type(pointer.dtype)
            elif name.startswith('decoder.sentence_encoder.layers.'):
                name = name.lstrip('decoder.sentence_encoder.layers.')
                layer_num, layer_name = name.split('.', 1)
                if layer_name in attention_tensor_map:
                    v_names = attention_tensor_map[layer_name]
                    if isinstance(v_names, str):
                        pointer = self._name_to_variable(
                            v_names.format(layer_num))
                        assert pointer.shape == tensor.shape
                        pointer.data = tensor.data.type(pointer.dtype)
                    else:
                        # Q, K, V in self-attention
                        tensors = torch.chunk(tensor, chunks=3, dim=0)
                        for i in range(3):
                            pointer = self._name_to_variable(
                                v_names[i].format(layer_num))
                            assert pointer.shape == tensors[i].shape
                            pointer.data = tensors[i].data.type(pointer.dtype)
                else:
                    raise NameError(f"Layer name '{layer_name}' not found")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for summary related operations.
"""
import unittest

import numpy as np

from texar.torch.run.metric.summary import *


class RegressionMetricTest(unittest.TestCase):
    def setUp(self) -> None:
        self.n_examples = 100
        self.batch_size = 2
        self.values = np.random.randn(self.n_examples)

    def test_running_average(self):
        queue_size = 10
        metric = RunningAverage(queue_size)
        for idx in range(0, self.n_examples, self.batch_size):
            end_idx = idx + self.batch_size
            metric.add(self.values[idx:end_idx], None)
            value = metric.value()
            answer = self.values[max(0, end_idx - queue_size):end_idx].mean()
            self.assertAlmostEqual(value, answer)
